#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .elastic import ElasticOcean

class PhabricatorOcean(ElasticOcean):
    """Phabricator Ocean feeder"""

    def _fix_item(self, item):
        item["ocean-unique-id"] = item["uuid"]
        # https://discuss.elastic.co/t/field-name-cannot-contain/33251/16
        if 'custom.external_reference' in item["data"]['fields']:
            item["data"]['fields']["custom_external_reference"] = item["data"]['fields'].pop("custom.external_reference")
        if 'custom.security_topic' in item["data"]['fields']:
            item["data"]['fields']["custom_security_topic"] = item["data"]['fields'].pop("custom.security_topic")

    def get_elastic_mappings(self):
        # This field data.transaction has string arrays and dicts arrays
        mapping = '''
         {
            "dynamic":true,
                "properties": {
                    "data": {
                        "properties": {
                            "transactions": {
                                "dynamic":false,
                                "properties": {}
                            }
                        }
                    }
                }
        }
        '''

        return {"items":mapping}



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Panels lib
#
# Copyright (C) 2016 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging

from os import sys

import requests

from .elk.elastic import ElasticSearch


def get_dashboard_json(elastic, dashboard):
    dash_json_url = elastic.index_url+"/dashboard/"+dashboard

    r = requests.get(dash_json_url, verify=False)

    dash_json = r.json()
    if "_source" not in dash_json:
        logging.error("Can not find dashboard: %s", dashboard)
        print (dash_json_url)
        sys.exit(1)

    return dash_json['_source']

def get_vis_json(elastic, vis):
    vis_json_url = elastic.index_url+"/visualization/"+vis

    r = requests.get(vis_json_url, verify=False)

    vis_json = r.json()
    if "_source" not in vis_json:
        logging.error("Can not find vis: %s (%s)", vis, vis_json_url)
        return

    return vis_json['_source']

def get_search_json(elastic, search_id):
    search_json_url = elastic.index_url+"/search/"+search_id

    r = requests.get(search_json_url, verify=False)

    search_json = r.json()
    if "_source" not in search_json:
        logging.error("Can not find search: %s (%s)", search_json_url)
        return
    return search_json['_source']

def get_index_pattern_json(elastic, index_pattern):
    index_pattern_json_url = elastic.index_url+"/index-pattern/"+index_pattern

    r = requests.get(index_pattern_json_url, verify=False)

    index_pattern_json = r.json()
    if "_source" not in index_pattern_json:
        logging.error("Can not find index_pattern_json: %s", index_pattern_json_url)
        return
    return index_pattern_json['_source']

def get_search_from_vis(elastic, vis):
    search_id = None
    vis_json = get_vis_json(elastic, vis)
    if not vis_json:
        search_id
    # The index pattern could be in search or in state
    # First search for it in saved search
    if "savedSearchId" in vis_json:
        search_id = vis_json["savedSearchId"]
    return search_id


def create_search(elastic_url, dashboard, index_pattern, es_index=None):
    """ Create the base search for vis if used

        :param elastic_url: URL for ElasticSearch (ES) server
        :param dashboard: kibana dashboard to be used as template
        :param enrich_index: ES index with enriched items used in the new dashboard

    """

    search_id = None
    if not es_index:
        es_index = ".kibana"
    elastic = ElasticSearch(elastic_url, es_index)

    dash_data = get_dashboard_json(elastic, dashboard)

    # First vis
    if "panelsJSON" not in dash_data:
        logging.error("Can not find vis in dashboard: %s", dashboard)
        raise

    # Get the search from the first vis in the panel
    for panel in json.loads(dash_data["panelsJSON"]):
        panel_id = panel["id"]
        logging.debug("Checking search in %s vis", panel_id)

        search_id = get_search_from_vis(elastic, panel_id)
        if search_id:
            break

    # And now time to create the search found
    if not search_id:
        logging.info("Can't find search  %s", dashboard)
        return

    logging.debug("Found template search %s", search_id)

    search_json = get_search_json(elastic, search_id)
    search_source = search_json['kibanaSavedObjectMeta']['searchSourceJSON']
    new_search_source = json.loads(search_source)
    new_search_source['index'] = index_pattern
    new_search_source = json.dumps(new_search_source)
    search_json['kibanaSavedObjectMeta']['searchSourceJSON'] = new_search_source

    search_json['title'] += " " + index_pattern
    new_search_id = search_id+"__"+index_pattern

    url = elastic.index_url+"/search/"+new_search_id
    requests.post(url, data = json.dumps(search_json), verify=False)

    logging.debug("New search created: %s", url)

    return new_search_id

def get_index_pattern_from_meta(meta_data):
    index = None
    mdata = meta_data["searchSourceJSON"]
    mdata = json.loads(mdata)
    if "index" in mdata:
        index = mdata["index"]
    if "filter" in mdata:
        if len(mdata["filter"]) > 0:
            index = mdata["filter"][0]["meta"]["index"]
    return index

def get_index_pattern_from_vis(elastic, vis):
    index_pattern = None
    vis_json = get_vis_json(elastic, vis)
    if not vis_json:
        return
    # The index pattern could be in search or in state
    # First search for it in saved search
    if "savedSearchId" in vis_json:
        search_json_url = elastic.index_url+"/search/"+vis_json["savedSearchId"]
        search_json = requests.get(search_json_url, verify=False).json()["_source"]
        index_pattern = get_index_pattern_from_meta(search_json["kibanaSavedObjectMeta"])
    elif "kibanaSavedObjectMeta" in vis_json:
        index_pattern = get_index_pattern_from_meta(vis_json["kibanaSavedObjectMeta"])
    return index_pattern


def create_index_pattern(elastic_url, dashboard, enrich_index, es_index=None):
    """ Create a index pattern using as template the index pattern
        in dashboard template vis

        :param elastic_url: URL for ElasticSearch (ES) server
        :param dashboard: kibana dashboard to be used as template
        :param enrich_index: ES index with enriched items used in the new dashboard

    """

    index_pattern = None
    if not es_index:
        es_index = ".kibana"
    elastic = ElasticSearch(elastic_url, es_index)

    dash_data = get_dashboard_json(elastic, dashboard)

    # First vis
    if "panelsJSON" not in dash_data:
        logging.error("Can not find vis in dashboard: %s", dashboard)
        raise

    # Get the index pattern from the first vis in the panel
    # that as index pattern data
    for panel in json.loads(dash_data["panelsJSON"]):
        panel_id = panel["id"]
        logging.debug("Checking index pattern in %s vis", panel_id)

        index_pattern = get_index_pattern_from_vis(elastic, panel_id)
        if index_pattern:
            break

    # And now time to create the index pattern found
    if not index_pattern:
        logging.error("Can't find index pattern for %s", dashboard)
        raise

    logging.debug("Found %s template index pattern", index_pattern)


    new_index_pattern_json = get_index_pattern_json(elastic, index_pattern)

    new_index_pattern_json['title'] = enrich_index
    url = elastic.index_url+"/index-pattern/"+enrich_index
    requests.post(url, data = json.dumps(new_index_pattern_json), verify=False)

    logging.debug("New index pattern created: %s", url)

    return enrich_index

def create_dashboard(elastic_url, dashboard, enrich_index, kibana_host, es_index=None):
    """ Create a new dashboard using dashboard as template
        and reading the data from enriched_index """

    def new_panels(elastic, panels, search_id):
        """ Create the new panels and their vis for the dashboard from the
            panels in the template dashboard """

        dash_vis_ids = []
        new_panels = []
        for panel in panels:
            if panel['type'] in ['visualization', 'search']:
                if panel['type'] == 'visualization':
                    dash_vis_ids.append(panel['id'])
                panel['id'] += "__"+enrich_index
                if panel['type'] == 'search':
                    panel['id'] = search_id
            new_panels.append(panel)

        create_vis(elastic, dash_vis_ids, search_id)


        return new_panels

    def create_vis(elastic, dash_vis_ids, search_id):
        """ Create new visualizations for the dashboard """

        # Create visualizations for the new dashboard
        item_template_url = elastic.index_url+"/visualization"
        # Hack: Get all vis if they are <10000. Use scroll API to get all.
        # Better: use mget to get all vis in dash_vis_ids
        item_template_url_search = item_template_url+"/_search?size=10000"
        r = requests.get(item_template_url_search, verify=False)
        all_visualizations =r.json()['hits']['hits']

        visualizations = []
        for vis in all_visualizations:
            if vis['_id'] in dash_vis_ids:
                visualizations.append(vis)

        logging.info("Total template vis found: %i", len(visualizations))

        for vis in visualizations:
            vis_data = vis['_source']
            vis_name = vis['_id'].split("_")[-1]
            vis_id = vis_name+"__"+enrich_index
            vis_data['title'] = vis_id
            vis_meta = json.loads(vis_data['kibanaSavedObjectMeta']['searchSourceJSON'])
            vis_meta['index'] = enrich_index
            vis_data['kibanaSavedObjectMeta']['searchSourceJSON'] = json.dumps(vis_meta)
            if "savedSearchId" in vis_data:
                vis_data["savedSearchId"] = search_id

            url = item_template_url+"/"+vis_id

            r = requests.post(url, data = json.dumps(vis_data), verify=False)
            logging.debug("Created new vis %s", url)

    if not es_index:
        es_index = ".kibana"

    # First create always the index pattern as data source
    index_pattern = create_index_pattern(elastic_url, dashboard, enrich_index, es_index)
    # If search is used create a new search with the new index_pàttern
    search_id = create_search(elastic_url, dashboard, index_pattern, es_index)

    elastic = ElasticSearch(elastic_url, es_index)

    # Create the new dashboard from the template
    dash_data = get_dashboard_json(elastic, dashboard)
    dash_data['title'] = enrich_index
    # Load template panels to create the new ones with their new vis
    panels = json.loads(dash_data['panelsJSON'])
    dash_data['panelsJSON'] = json.dumps(new_panels(elastic, panels, search_id))
    dash_path = "/dashboard/"+dashboard+"__"+enrich_index
    url = elastic.index_url + dash_path
    requests.post(url, data = json.dumps(dash_data), verify=False)

    dash_url = kibana_host+"/app/kibana#"+dash_path
    return dash_url

def list_dashboards(elastic_url, es_index=None):
    if not es_index:
        es_index = ".kibana"

    elastic = ElasticSearch(elastic_url, es_index)

    dash_json_url = elastic.index_url+"/dashboard/_search?size=10000"

    print (dash_json_url)

    r = requests.get(dash_json_url, verify=False)

    res_json = r.json()

    if "hits" not in res_json:
        logging.error("Can't find dashboards")
        raise RuntimeError("Can't find dashboards")

    for dash in res_json["hits"]["hits"]:
        print (dash["_id"])

def get_dashboard_name(import_file):
    """ Return the dashboard name included in a JSON file """

    name = None

    with open(import_file, 'r') as f:
        try:
            kibana = json.loads(f.read())
        except ValueError:
            logging.error("Wrong file format")

        if 'dashboard' not in kibana:
            logging.error("Wrong file format. Can't find 'dashboard' field.")
        name = kibana['dashboard']['id']

    return name

def import_dashboard(elastic_url, import_file, es_index=None):
    logging.debug("Reading from %s the JSON for the dashboard to be imported",
                  import_file)

    with open(import_file, 'r') as f:
        try:
            kibana = json.loads(f.read())
        except ValueError:
            logging.error("Wrong file format")
            sys.exit(1)

        if 'dashboard' not in kibana:
            logging.error("Wrong file format. Can't find 'dashboard' field.")
            sys.exit(1)

        if not es_index:
            es_index = ".kibana"
        elastic = ElasticSearch(elastic_url, es_index)

        url = elastic.index_url+"/dashboard/"+kibana['dashboard']['id']
        requests.post(url, data = json.dumps(kibana['dashboard']['value']), verify=False)

        if 'searches' in kibana:
            for search in kibana['searches']:
                url = elastic.index_url+"/search/"+search['id']
                requests.post(url, data = json.dumps(search['value']), verify=False)

        if 'index_patterns' in kibana:
            for index in kibana['index_patterns']:
                url = elastic.index_url+"/index-pattern/"+index['id']
                requests.post(url, data = json.dumps(index['value']), verify=False)

        if 'visualizations' in kibana:
            for vis in kibana['visualizations']:
                url = elastic.index_url+"/visualization"+"/"+vis['id']
                requests.post(url, data = json.dumps(vis['value']), verify=False)

        logging.debug("Done")


def export_dashboard(elastic_url, dash_id, export_file, es_index=None):

    # Kibana dashboard fields
    kibana = {"dashboard": None,
              "visualizations": [],
              "index_patterns": [],
              "searches": []}

    # Used to avoid having duplicates
    search_ids_done = []
    index_ids_done = []

    logging.debug("Exporting dashboard %s to %s", dash_id, export_file)
    if not es_index:
        es_index = ".kibana"

    elastic = ElasticSearch(elastic_url, es_index)

    kibana["dashboard"] = {"id":dash_id, "value":get_dashboard_json(elastic, dash_id)}

    if "panelsJSON" not in kibana["dashboard"]["value"]:
        # The dashboard is empty. No visualizations included.
        return kibana

    # Export all visualizations and the index patterns and searches in them
    for panel in json.loads(kibana["dashboard"]["value"]["panelsJSON"]):
        if panel['type'] in ['visualization']:
            vis_id = panel['id']
            vis_json = get_vis_json(elastic, vis_id)
            kibana["visualizations"].append({"id": vis_id, "value": vis_json})
            search_id = get_search_from_vis(elastic, vis_id)
            if search_id and search_id not in search_ids_done:
                search_ids_done.append(search_id)
                kibana["searches"].append({"id":search_id,
                                           "value":get_search_json(elastic, search_id)})
            index_pattern_id = get_index_pattern_from_vis(elastic, vis_id)
            if index_pattern_id and index_pattern_id not in index_ids_done:
                index_ids_done.append(index_pattern_id)
                kibana["index_patterns"].append({"id":index_pattern_id,
                                                 "value":get_index_pattern_json(elastic, index_pattern_id)})
    logging.debug("Done")

    with open(export_file, 'w') as f:
        f.write(json.dumps(kibana, indent=4, sort_keys=True))



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging

from dateutil import parser

from .utils import get_time_diff_days, unixtime_to_datetime

from .enrich import Enrich, metadata

class AskbotEnrich(Enrich):

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        # question
        user = self.get_sh_identity(item, self.get_field_author())
        identities.append(user)

        # answers
        if 'answers' in item['data']:
            for answer in item['data']['answers']:
                # avoid "answered_by" : "This post is a wiki" corner case
                if type(answer['answered_by']) is dict:
                    user = self.get_sh_identity(answer['answered_by'])
                    identities.append(user)
                if 'comments' in answer:
                    for comment in answer['comments']:
                        identities.append(self.get_sh_identity(comment))
        return identities

    def get_sh_identity(self, item, identity_field=None):
        identity = {key: None for key in ['username', 'name', 'email']}

        user = item  # by default a specific user dict is expected
        if 'data' in item and type(item) == dict:
            user = item['data'][identity_field]
        elif 'author' in item:
            user = item['author']

        if user is None:
            return identity

        if 'username' in user:
            identity['username'] = user['username']
        elif 'user_display_name' in user:
            identity['username'] = user['user_display_name']
        identity['name'] = identity['username']
        return identity

    def get_field_author(self):
        return 'author'

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "author_badges": {
                  "type": "string",
                  "index":"analyzed"
                },
                "summary": {
                  "type": "string",
                  "index":"analyzed"
                },
                "question_tags": {
                    "type": "string",
                    "index":"analyzed",
                    "analyzer" : "comma"
                },
                "question_answer_ids": {
                    "type": "string",
                    "index":"analyzed",
                    "analyzer" : "comma"
                }
           }
        } """
        return {"items":mapping}

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        copy_fields = ["metadata__updated_on", "metadata__timestamp",
                       "ocean-unique-id", "origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None

        question = item['data']

        # Fields that are the same in item and eitem
        copy_fields = ["id", "url", "title", "summary", "score"]
        for f in copy_fields:
            if f in question:
                eitem[f] = question[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"title": "question_title",
                      "answer_count": "question_answer_count",
                      "view_count": "question_view_count",
                      "answer_ids": "question_answer_ids"}
        for fn in map_fields:
            if fn in question:
                eitem[map_fields[fn]] = question[fn]
            else:
                eitem[map_fields[fn]] = None

        # First answer time
        added_at = unixtime_to_datetime(float(question["added_at"]))
        eitem['time_to_reply'] = None
        if 'answers' in question:
            # answers ordered by time
            first_answer_time = unixtime_to_datetime(float(question['answers'][0]["added_at"]))
            eitem['time_to_reply'] = get_time_diff_days(added_at, first_answer_time)

        if question['author'] and type(question['author']) is dict:
            eitem['author_user_name'] = question['author']['username']
            eitem['author_id'] = question['author']['id']
            eitem['author_badges'] = question['author']['badges']
            eitem['author_reputation'] = int(question['author']['reputation'])

        eitem['question_last_activity_at'] = unixtime_to_datetime(float(question['last_activity_at'])).isoformat()
        eitem['question_last_activity_by_id'] = question['last_activity_by']['id']
        eitem['question_last_activity_by_username'] = question['last_activity_by']['username']
        # Analyzed
        eitem['question_tags'] = ",".join([tag for tag in question['tags']])
        eitem['question_answer_ids'] = ",".join([str(aid) for aid in question['answer_ids']])

        eitem['comment_count'] = 0
        if 'answers' in question:
            eitem['comment_count'] = sum([len (a['comments']) if 'comments' in a else 0 for a in question['answers']])

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))
        eitem["type"] = "question"
        eitem.update(self.get_grimoire_fields(added_at.isoformat(), eitem["type"]))

        return eitem

    def get_field_unique_id_comment(self):
        return "id"

    def get_field_unique_id_answer(self):
        return "id"

    def get_rich_comment(self, item, answer, comment):
        ecomment = self.get_rich_item(item)  # reuse all fields from item
        ecomment['id'] = comment['id']
        ecomment['url'] = item['data']['url']+"/?answer="
        ecomment['url'] += answer['id']+'#post-id-'+answer['id']
        if 'author' in comment:
            ecomment['author_user_name'] = comment['author']['username']
            ecomment['author_id'] = comment['author']['id']
        if 'summary' in comment:
            ecomment['summary'] = comment['summary']
        ecomment['score'] = comment['score']

        dfield = 'added_at'
        if 'comment_added_at' in comment:
            dfield = 'comment_added_at'

        if self.sortinghat:
            if dfield == 'added_at':
                comment['added_at_date'] = unixtime_to_datetime(float(comment[dfield])).isoformat()
            else:
                comment['added_at_date'] = comment[dfield]
            ecomment.update(self.get_item_sh(comment, date_field="added_at_date"))

        if dfield == 'added_at':
            comment_at = unixtime_to_datetime(float(comment[dfield]))
        else:
            comment_at = parser.parse(comment[dfield])

        added_at = unixtime_to_datetime(float(item['data']["added_at"]))
        ecomment['time_from_question'] = get_time_diff_days(added_at, comment_at)
        ecomment['type'] = 'comment'
        ecomment.update(self.get_grimoire_fields(comment_at.isoformat(), ecomment['type']))

        # Clean items fields not valid in comments
        for f in ['is_askbot_question', 'author_reputation', 'author_badges', 'is_correct', 'comment_count']:
            if f in ecomment:
                ecomment.pop(f)

        return ecomment

    def get_rich_answer(self, item, answer):
        eanswer = self.get_rich_item(item)  # reuse all fields from item
        eanswer['id'] = answer['id']
        eanswer['url'] = item['data']['url']+"/?answer="
        eanswer['url'] += answer['id']+'#post-id-'+answer['id']
        if type(answer['answered_by']) is dict:
            eanswer['author_user_name'] = answer['answered_by']['username']
            eanswer['author_id'] = answer['answered_by']['id']
            eanswer['author_badges'] = answer['answered_by']['badges']
            eanswer['author_reputation'] = int(answer['answered_by']['reputation'])
        eanswer['summary'] = answer['summary']
        eanswer['score'] = answer['score']
        if 'is_correct' in answer:
            eanswer['is_correct'] = 1

        if self.sortinghat:
            answer['added_at_date'] = unixtime_to_datetime(float(answer["added_at"])).isoformat()
            eanswer.update(self.get_item_sh(answer, date_field="added_at_date"))

        answer_at = unixtime_to_datetime(float(answer["added_at"]))
        added_at = unixtime_to_datetime(float(item['data']["added_at"]))
        eanswer['time_from_question'] = get_time_diff_days(added_at, answer_at)
        eanswer['type'] = 'answer'
        eanswer.update(self.get_grimoire_fields(answer_at.isoformat(), eanswer['type']))

        # Clean items fields not valid in comments
        eanswer.pop('is_askbot_question')

        return eanswer

    def get_rich_item_answers_comments(self, item):
        answers_enrich = []
        comments_enrich = []

        if 'answers' not in item['data']:
            return (answers_enrich, comments_enrich)

        for answer in item['data']['answers']:
            eanswer = self.get_rich_answer(item, answer)
            if not answers_enrich:
                eanswer['first_answer'] = 1
            answers_enrich.append(eanswer)

            # And now time to process the comments
            if 'comments' in answer:
                for comment in answer['comments']:
                    ecomment = self.get_rich_comment(item, answer, comment)
                    comments_enrich.append(ecomment)

        return (answers_enrich, comments_enrich)

    def enrich_items(self, items):
        nitems = super(AskbotEnrich, self).enrich_items(items)
        logging.info("Total questions enriched: %i", nitems)

        # And now for each item we want also the answers (tops)
        nanswers = 0
        ncomments = 0
        rich_item_answers = []
        rich_item_comments = []

        for item in items:
            (answers, comments) = self.get_rich_item_answers_comments(item)
            rich_item_answers += answers
            rich_item_comments += comments

        if rich_item_answers:
            nanswers = self.elastic.bulk_upload(rich_item_answers,
                                                self.get_field_unique_id_answer())
        if rich_item_comments:
            ncomments += self.elastic.bulk_upload(rich_item_comments,
                                                  self.get_field_unique_id_comment())
        logging.info("Total answers enriched: %i", nanswers)
        logging.info("Total comments enriched: %i", ncomments)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from .enrich import Enrich, metadata

class ReMoEnrich(Enrich):

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)
        self.author = "user"  # changes if we are generating events

    def get_field_author(self):
        return self.author

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "description_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                  },
                "geolocation": {
                   "type": "geo_point"
                },
                "functional_areas": {
                  "type": "string",
                  "analyzer": "comma"
                  },
                "categories": {
                  "type": "string",
                  "analyzer": "comma"
                  }
           }
        } """

        return {"items":mapping}

    def get_identities(self, item):
        ''' Return the identities from an item '''

        identities = []
        item = item['data']
        if 'owner' in item:
            identities.append(self.get_sh_identity(item['owner']))
        if 'user' in item:
            identities.append(self.get_sh_identity(item['user']))
        if 'mentor' in item:
            identities.append(self.get_sh_identity(item['mentor']))

        return identities

    def get_sh_identity(self, item, identity_field=None):
		# "owner": {
        #             "first_name": "Huda",
        #             "last_name": "Sarfraz",
        #             "_url": "https://reps.mozilla.org/api/beta/users/959/",
        #             "display_name": "huda_sarfraz"
        #          },
        identity = {'username':None, 'email':None, 'name':None}

        if not item:
            return identity

        user = item
        if 'data' in item and type(item) == dict:
            user = item['data'][identity_field]

        identity['username'] = user["display_name"]
        identity['email'] = None
        identity['name'] = user["first_name"]+" "+user["last_name"]

        return identity

    def get_item_category(self, item):
        # We need to detect the category of item: activities (report), events or users
        category = None

        item = item['data']

        if 'estimated_attendance' in item:
            category = 'events'
        elif 'activity' in item:
            category = 'activities'
        elif 'first_name' in item:
            category = 'users'
        else:
            logging.error("Can not detect category in item %s", item)

        return category

    @metadata
    def get_rich_item(self, item):
        # We need to detect the category of item: activities (report), events or users
        eitem = {}

        eitem.update(self.__get_rich_item_common(item))

        category = self.get_item_category(item)

        if category == 'activities':
            eitem.update(self.__get_rich_item_activities(item))
        elif category == 'users':
            eitem.update(self.__get_rich_item_users(item))
        elif category == 'events':
            eitem.update(self.__get_rich_item_events(item))

        return eitem

    def __get_rich_item_common(self, item):
        # metadata fields to copy

        eitem = {}

        copy_fields = ["metadata__updated_on", "metadata__timestamp",
                       "ocean-unique-id", "origin", "offset"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None

        return eitem


    def __get_rich_item_activities(self, item):
        eitem = {}

        # The real data
        activity = item['data']


        # data fields to copy
        copy_fields = ["initiative","location","activity","link",
                       "activity_description","remo_url",
                       "external_link","latitude","longitude", "report_date"]
        for f in copy_fields:
            if f in activity:
                eitem[f] = activity[f]
            else:
                eitem[f] = None


        eitem['user_profile_url'] = activity['user']['_url']
        eitem['user'] = activity['user']['first_name']+" "+activity['user']['last_name']

        if 'mentor' in activity and activity['mentor']:
            eitem['mentor_profile_url'] = activity['mentor']['_url']
            eitem['mentor'] = activity['mentor']['first_name']+" "+activity['mentor']['last_name']

        # geolocation
        if -90<int(eitem['latitude'])<90 and \
            -180<int(eitem['longitude'])<180:
            eitem['geolocation'] = {
                "lat": eitem['latitude'],
                "lon": eitem['longitude'],
            }

        eitem['functional_areas'] = ''
        for area in eitem['functional_areas']:
            eitem['functional_areas'] += "," + area['name']

        if self.sortinghat:
            self.author = "user"
            eitem.update(self.get_item_sh(item))

        return eitem

    def __get_rich_item_users(self, item):
        eitem = {}

        return eitem

    def __get_rich_item_events(self, item):
        eitem = {}

        # The real data
        event = item['data']

        # data fields to copy
        copy_fields = ["initiative","city","country",
                       "description","estimated_attendance","remo_url",
                       "external_link","lat","lon", "region","timezone",
                       "planning_pad_url","hashtag"]
        for f in copy_fields:
            if f in event:
                eitem[f] = event[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {
            "description": "description_analyzed",
            "end":"end_date",
            "start":"start_date",
            "name":"title"
        }
        for fn in map_fields:
            eitem[map_fields[fn]] = event[fn]

        eitem['owner_profile_url'] = None
        eitem['owner'] = None
        if 'owner' in event:
            eitem['owner_profile_url'] = event['owner']['_url']
            eitem['owner'] = event['owner']['display_name']

        # geolocation
        if -90<int(eitem['lat'])<90 and \
            -180<int(eitem['lon'])<180:
            eitem['geolocation'] = {
                "lat": eitem['lat'],
                "lon": eitem['lon'],
            }

        eitem['categories'] = ''
        for cat in event['categories']:
            eitem['categories'] += "," + cat['name']
        eitem['categories'] = eitem['categories'][1:]  # remove first comma

        if self.sortinghat:
            self.author = 'owner'
            eitem.update(self.get_item_sh(item))

        return eitem



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Grimoire general utils
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import argparse
from dateutil import parser
import logging
import sys

from .ocean.elastic import ElasticOcean

# Connectors for Ocean
from .ocean.askbot import AskbotOcean
from .ocean.bugzilla import BugzillaOcean
from .ocean.bugzillarest import BugzillaRESTOcean
from .ocean.confluence import ConfluenceOcean
from .ocean.discourse import DiscourseOcean
from .ocean.gerrit import GerritOcean
from .ocean.git import GitOcean
from .ocean.github import GitHubOcean
from .ocean.jenkins import JenkinsOcean
from .ocean.jira import JiraOcean
from .ocean.kitsune import KitsuneOcean
from .ocean.mbox import MBoxOcean
from .ocean.mediawiki import MediaWikiOcean
from .ocean.meetup import MeetupOcean
from .ocean.phabricator import PhabricatorOcean
from .ocean.redmine import RedmineOcean
from .ocean.remo2 import ReMoOcean
from .ocean.rss import RSSOcean
from .ocean.stackexchange import StackExchangeOcean
from .ocean.supybot import SupybotOcean
from .ocean.telegram import TelegramOcean
from .ocean.twitter import TwitterOcean

# Connectors for EnrichOcean
from .elk.askbot import AskbotEnrich
from .elk.bugzilla import BugzillaEnrich
from .elk.bugzillarest import BugzillaRESTEnrich
from .elk.confluence import ConfluenceEnrich
from .elk.discourse import DiscourseEnrich
from .elk.git import GitEnrich
from .elk.github import GitHubEnrich
from .elk.gerrit import GerritEnrich
from .elk.gmane import GmaneEnrich
from .elk.jenkins import JenkinsEnrich
from .elk.jira import JiraEnrich
from .elk.kitsune import KitsuneEnrich
from .elk.mbox import MBoxEnrich
from .elk.mediawiki import MediaWikiEnrich
from .elk.meetup import MeetupEnrich
from .elk.phabricator import PhabricatorEnrich
from .elk.redmine import RedmineEnrich
# from .elk.remo import ReMoEnrich
from .elk.remo2 import ReMoEnrich
from .elk.rss import RSSEnrich
from .elk.pipermail import PipermailEnrich
from .elk.stackexchange import StackExchangeEnrich
from .elk.supybot import SupybotEnrich
from .elk.telegram import TelegramEnrich
from .elk.twitter import TwitterEnrich

# Connectors for Perceval
from perceval.backends.core.askbot import Askbot, AskbotCommand
from perceval.backends.core.bugzilla import Bugzilla, BugzillaCommand
from perceval.backends.core.bugzillarest import BugzillaREST, BugzillaRESTCommand
from perceval.backends.core.discourse import Discourse, DiscourseCommand
from perceval.backends.core.confluence import Confluence, ConfluenceCommand
from perceval.backends.core.gerrit import Gerrit, GerritCommand
from perceval.backends.core.git import Git, GitCommand
from perceval.backends.core.github import GitHub, GitHubCommand
from perceval.backends.core.gmane import Gmane, GmaneCommand
from perceval.backends.core.jenkins import Jenkins, JenkinsCommand
from perceval.backends.core.jira import Jira, JiraCommand
from perceval.backends.mozilla.kitsune import Kitsune, KitsuneCommand
from perceval.backends.core.mbox import MBox, MBoxCommand
from perceval.backends.core.mediawiki import MediaWiki, MediaWikiCommand
from perceval.backends.core.meetup import Meetup, MeetupCommand
from perceval.backends.core.phabricator import Phabricator, PhabricatorCommand
from perceval.backends.core.pipermail import Pipermail, PipermailCommand
from perceval.backends.core.redmine import Redmine, RedmineCommand
from perceval.backends.mozilla.remo import ReMo, ReMoCommand
from perceval.backends.core.rss import RSS, RSSCommand
from perceval.backends.core.stackexchange import StackExchange, StackExchangeCommand
from perceval.backends.core.supybot import Supybot, SupybotCommand
from perceval.backends.core.telegram import Telegram, TelegramCommand


from .elk.elastic import ElasticSearch
from .elk.elastic import ElasticConnectException

def get_connector_from_name(name):
    found = None
    connectors = get_connectors()

    for cname in connectors:
        if cname == name:
            found = connectors[cname]

    return found

def get_connector_name(cls):
    found = None
    connectors = get_connectors()

    for cname in connectors:
        for con in connectors[cname]:
            if cls == con:
                if found:
                    # The canonical name is included in the classname
                    if cname in cls.__name__.lower():
                        found = cname
                else:
                    found = cname
    return found

def get_connectors():

    return {"askbot":[Askbot, AskbotOcean, AskbotEnrich, AskbotCommand],
            "bugzilla":[Bugzilla, BugzillaOcean, BugzillaEnrich, BugzillaCommand],
            "bugzillarest":[BugzillaREST, BugzillaRESTOcean, BugzillaRESTEnrich, BugzillaRESTCommand],
            "confluence":[Confluence, ConfluenceOcean, ConfluenceEnrich, ConfluenceCommand],
            "discourse":[Discourse, DiscourseOcean, DiscourseEnrich, DiscourseCommand],
            "gerrit":[Gerrit, GerritOcean, GerritEnrich, GerritCommand],
            "git":[Git, GitOcean, GitEnrich, GitCommand],
            "github":[GitHub, GitHubOcean, GitHubEnrich, GitHubCommand],
            "gmane":[Gmane, MBoxOcean, GmaneEnrich, GmaneCommand],
            "jenkins":[Jenkins, JenkinsOcean, JenkinsEnrich, JenkinsCommand],
            "jira":[Jira, JiraOcean, JiraEnrich, JiraCommand],
            "kitsune":[Kitsune, KitsuneOcean, KitsuneEnrich, KitsuneCommand],
            "mbox":[MBox, MBoxOcean, MBoxEnrich, MBoxCommand],
            "mediawiki":[MediaWiki, MediaWikiOcean, MediaWikiEnrich, MediaWikiCommand],
            "meetup":[Meetup, MeetupOcean, MeetupEnrich, MeetupCommand],
            "phabricator":[Phabricator, PhabricatorOcean, PhabricatorEnrich, PhabricatorCommand],
            "pipermail":[Pipermail, MBoxOcean, MBoxEnrich, PipermailCommand],
            "pipermail":[Pipermail, MBoxOcean, PipermailEnrich, PipermailCommand],
            "redmine":[Redmine, RedmineOcean, RedmineEnrich, RedmineCommand],
            "remo":[ReMo, ReMoOcean, ReMoEnrich, ReMoCommand],
            "rss":[RSS, RSSOcean, RSSEnrich, RSSCommand],
            "stackexchange":[StackExchange, StackExchangeOcean,
                             StackExchangeEnrich, StackExchangeCommand],
             "supybot":[Supybot, SupybotOcean, SupybotEnrich, SupybotCommand],
             "telegram":[Telegram, TelegramOcean, TelegramEnrich, TelegramCommand],
             "twitter":[None, TwitterOcean, TwitterEnrich, None]
            }  # Will come from Registry

def get_elastic(url, es_index, clean = None, backend = None):

    mapping = None

    if backend:
        mapping = backend.get_elastic_mappings()
        analyzers = backend.get_elastic_analyzers()
    try:
        insecure = True
        elastic = ElasticSearch(url, es_index, mapping, clean, insecure, analyzers)

    except ElasticConnectException:
        logging.error("Can't connect to Elastic Search. Is it running?")
        sys.exit(1)

    return elastic

def config_logging(debug):

    if debug:
        logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(message)s')
        logging.debug("Debug mode activated")
    else:
        logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)

ARTHUR_USAGE_MSG = ''
ARTHUR_DESC_MSG = ''
ARTHUR_EPILOG_MSG = ''

def get_params_parser():
    """Parse command line arguments"""

    parser = argparse.ArgumentParser(usage=ARTHUR_USAGE_MSG,
                                     description=ARTHUR_DESC_MSG,
                                     epilog=ARTHUR_EPILOG_MSG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     add_help=False)

    ElasticOcean.add_params(parser)

    parser.add_argument('-h', '--help', action='help',
                       help=argparse.SUPPRESS)
    parser.add_argument('-g', '--debug', dest='debug',
                        action='store_true',
                        help=argparse.SUPPRESS)

    parser.add_argument("--no_incremental",  action='store_true',
                        help="don't use last state for data source")
    parser.add_argument("--fetch_cache",  action='store_true',
                        help="Use cache for item retrieval")
    parser.add_argument("--redis",  default="redis",
                        help="url for the redis server")
    parser.add_argument("--enrich",  action='store_true',
                        help="Enrich items after retrieving")
    parser.add_argument("--enrich_only",  action='store_true',
                        help="Only enrich items (DEPRECATED, use --only-enrich)")
    parser.add_argument("--only-enrich",  dest='enrich_only', action='store_true',
                        help="Only enrich items (DEPRECATED, use --only-enrich)")
    parser.add_argument("--filter-raw",  dest='filter_raw',
                        help="Filter raw items. Format: field:value")
    parser.add_argument("--events-enrich",  dest='events_enrich', action='store_true',
                        help="Enrich events in items")
    parser.add_argument('--index', help="Ocean index name")
    parser.add_argument('--index-enrich', dest="index_enrich", help="Ocean enriched index name")
    parser.add_argument('--db-user', help="User for db connection (default to root)",
                        default="root")
    parser.add_argument('--db-password', help="Password for db connection (default empty)",
                        default="")
    parser.add_argument('--db-host', help="Host for db connection (default to mariadb)",
                        default="mariadb")
    parser.add_argument('--db-projects-map', help="Projects Mapping DB")
    parser.add_argument('--json-projects-map', help="Projects Mapping JSON file")
    parser.add_argument('--project', help="Project for the repository (origin)")
    parser.add_argument('--refresh-projects', action='store_true', help="Refresh projects in enriched items")
    parser.add_argument('--db-sortinghat', help="SortingHat DB")
    parser.add_argument('--only-identities', action='store_true', help="Only add identities to SortingHat DB")
    parser.add_argument('--refresh-identities', action='store_true', help="Refresh identities in enriched items")
    parser.add_argument('--author_id', help="Field author_id to be refreshed")
    parser.add_argument('--author_uuid', help="Field author_uuid to be refreshed")
    parser.add_argument('--github-token', help="If provided, github usernames will be retrieved in git enrich.")
    parser.add_argument('--studies', action='store_true', help="Execute studies after enrichment.")
    parser.add_argument('--only-studies', action='store_true', help="Execute only studies.")
    parser.add_argument('backend', help=argparse.SUPPRESS)
    parser.add_argument('backend_args', nargs=argparse.REMAINDER,
                        help=argparse.SUPPRESS)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    return parser


def get_params():
    ''' Get params definition from ElasticOcean and from all the backends '''

    parser = get_params_parser()
    args = parser.parse_args()

    return args

def get_time_diff_days(start_txt, end_txt):
    ''' Number of days between two days  '''

    if start_txt is None or end_txt is None:
        return None

    start = parser.parse(start_txt)
    end = parser.parse(end_txt)

    seconds_day = float(60*60*24)
    diff_days = \
        (end-start).total_seconds() / seconds_day
    diff_days = float('%.2f' % diff_days)

    return diff_days



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# SortingHat class helper
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from datetime import datetime
import logging
import traceback

from sortinghat import api
from sortinghat.db.model import Identity, UniqueIdentity
from sortinghat.exceptions import AlreadyExistsError, NotFoundError, WrappedValueError
from sortinghat.matcher import create_identity_matcher

logger = logging.getLogger(__name__)

class SortingHat(object):

    @classmethod
    def get_uuid_from_id(cls, db, sh_id):
        uuid = None

        with db.connect() as session:
            query = session.query(Identity).\
            filter(Identity.id == sh_id)
            identities = query.all()
            if identities:
                uuid = identities[0].uuid
        return uuid

    @classmethod
    def get_github_commit_username(cls, db, identity, source):
        user = None

        with db.connect() as session:
            query = session.query(Identity).\
            filter(Identity.name == identity['name'], Identity.email == identity['email'], Identity.source == source)
            identities = query.all()
            if identities:
                user = {}
                user['name'] = identities[0].name
                user['email'] = identities[0].email
                user['username'] = identities[0].username
        return user

    @classmethod
    def add_identity(cls, db, identity, backend):
        """ Load and identity list from backend in Sorting Hat """
        uuid = None

        try:
            uuid = api.add_identity(db, backend, identity['email'],
                                    identity['name'], identity['username'])

            logger.debug("New sortinghat identity %s %s,%s,%s ",
                        uuid, identity['username'], identity['name'], identity['email'])

            profile = {"name": identity['name'] if identity['name'] else identity['username'],
                       "email": identity['email']}

            api.edit_profile(db, uuid, **profile)

        except AlreadyExistsError as ex:
            uuid = ex.uuid
        except WrappedValueError as ex:
            logging.warning("Trying to add a None identity. Ignoring it.")
        except UnicodeEncodeError as ex:
            logging.warning("UnicodeEncodeError. Ignoring it. %s %s %s",
                            identity['email'], identity['name'],
                            identity['username'])
        except Exception as ex:
            logging.warning("Unknown exception adding identity. Ignoring it. %s %s %s",
                            identity['email'], identity['name'],
                            identity['username'])
            traceback.print_exc()

        if 'company' in identity and identity['company'] is not None:
            try:
                api.add_organization(db, identity['company'])
                api.add_enrollment(db, uuid, identity['company'],
                                   datetime(1900, 1, 1),
                                   datetime(2100, 1, 1))
            except AlreadyExistsError:
                pass

        return uuid


    @classmethod
    def add_identities(cls, db, identities, backend):
        """ Load identities list from backend in Sorting Hat """

        logger.info("Adding the identities to SortingHat")

        total = 0

        for identity in identities:
            cls.add_identity(db, identity, backend)
            total +=1

        logging.info("Total identities added to SH: %i", total)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#


import logging

from .enrich import Enrich, metadata

from .utils import unixtime_to_datetime

class MeetupEnrich(Enrich):


    def get_field_author(self):
        return "author"

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "description_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                  },
                "geolocation": {
                   "type": "geo_point"
                },
                "group_geolocation": {
                   "type": "geo_point"
                },
                "group_topics" : {
                    "type" : "string",
                    "analyzer" : "comma"
                },
                "group_topics_keys" : {
                    "type" : "string",
                    "analyzer" : "comma"
                }
           }
        } """

        return {"items":mapping}

    def get_identities(self, item):
        ''' Return the identities from an item '''

        identities = []
        item = item['data']

        identities = []

        # Creators
        if 'event_hosts' in item:
            user = self.get_sh_identity(item['event_hosts'][0])
            identities.append(user)

        # rsvps
        for rsvp in item['rsvps']:
            user = self.get_sh_identity(rsvp['member'])
            identities.append(user)
        # Comments
        for comment in item['comments']:
            user = self.get_sh_identity(comment['member'])
            identities.append(user)
        return identities


    def get_sh_identity(self, item, identity_field=None):
        identity = {'username':None, 'email':None, 'name':None}

        if not item:
            return identity

        user = item
        if 'data' in item and type(item) == dict:
            user = item['data'][identity_field]

        identity['username'] = user["id"]
        identity['email'] = None
        identity['name'] = user["name"]

        return identity

    @metadata
    def get_rich_item(self, item):
        # We need to detect the category of item: activities (report), events or users
        eitem = {}

        copy_fields = ["metadata__updated_on", "metadata__timestamp",
                       "ocean-unique-id", "origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None

        event = item['data']

        # data fields to copy
        copy_fields = ["id", "how_to_find_us"]
        for f in copy_fields:
            if f in event:
                eitem[f] = event[f]
            else:
                eitem[f] = None

        # Fields which names are translated
        map_fields = {
            "link": "url",
            "rsvp_limit": "rsvps_limit"
        }
        for fn in map_fields:
            if fn in event:
                eitem[map_fields[fn]] = event[fn]
            else:
                eitem[f] = None


        # event host fields: author of the event
        if 'event_hosts' in event:
            host = event['event_hosts'][0]
            if 'photo' in host:
                eitem['member_photo_url'] = host['photo']['photo_link']
                eitem['member_photo_id'] = host['photo']['id']
                eitem['member_photo_type'] = host['photo']['type']
            eitem['member_is_host'] = True
            eitem['member_id'] = host['id']
            eitem['member_name'] = host['name']
            eitem['member_url'] = "https://www.meetup.com/members/" + str(host['id'])

        eitem['event_url'] = event['link']

        # data fields to copy with meetup`prefix
        copy_fields = ["description", "plain_text_description",
                       "created", "name", "status",
                       "time", "updated", "utc_offset", "visibility",
                       "waitlist_count", "yes_rsvp_count", "duration",
                       "featured", "rsvpable"]
        for f in copy_fields:
            if f in event:
                eitem["meetup_"+f] = event[f]
            else:
                eitem[f] = None

        eitem['num_rsvps'] = len(event['rsvps'])
        eitem['num_comments'] = len(event['comments'])

        if 'venue' in event:
            venue = event['venue']
            copy_fields = ["id", "name", "city", "state", "zip", "country",
                           "localized_country_name", "repinned", "address_1"]
            for f in copy_fields:
                if f in venue:
                    eitem["venue_"+f] = venue[f]
                else:
                    eitem[f] = None

            eitem['venue_geolocation'] = {
                "lat": event['venue']['lat'],
                "lon": event['venue']['lon'],
            }

        if 'series' in event:
            eitem['series_id'] = event['series']['id']
            eitem['series_description'] = event['series']['description']
            eitem['series_start_date'] = event['series']['start_date']


        if 'group' in event:
            group = event['group']
            copy_fields = ["id", "created", "join_mode", "name", "url_name",
                           "who"]
            for f in copy_fields:
                if f in group:
                    eitem["group_"+f] = group[f]
                else:
                    eitem[f] = None

            eitem['group_geolocation'] = {
                "lat": group['lat'],
                "lon": group['lon'],
            }
            group_topics = [topic['name'] for topic in group['topics']]
            group_topics_keys = [topic['urlkey'] for topic in group['topics']]
            eitem['group_topics'] = ",".join(group_topics)
            eitem['group_topics_keys'] = ",".join(group_topics_keys)

        if len(event['rsvps']) > 0:
            eitem['group_members'] = event['rsvps'][0]['group']['members']

        created = unixtime_to_datetime(event['created']/1000).isoformat()
        eitem['type'] = "meetup"
        eitem.update(self.get_grimoire_fields(created, eitem['type']))

        if self.sortinghat:
            eitem.update(self.get_item_sh(event))

        return eitem


    def get_item_sh(self, item):
        """ Add sorting hat enrichment fields  """

        sh_fields = {}

        # Not shared common get_item_sh because it is pretty specific
        if 'member' in item:
            # comment and rsvp
            identity  = self.get_sh_identity(item['member'])
        elif 'event_hosts' in item:
            # meetup event
            identity  = self.get_sh_identity(item['event_hosts'][0])
        else:
            return sh_fields

        created = unixtime_to_datetime(item['created']/1000)
        sh_fields = self.get_item_sh_fields(identity, created)

        return sh_fields

    def get_rich_item_comments(self, item):
        comments_enrich = []
        if 'comments' in item['data']:
            for comment in item['data']['comments']:
                ecomment = self.get_rich_item(item)  # reuse all fields from item
                created = unixtime_to_datetime(comment['created']/1000).isoformat()
                ecomment['url'] = comment['link']
                ecomment['id'] = comment['id']
                ecomment['comment'] = comment['comment']
                ecomment['like_count'] = comment['like_count']
                ecomment['type'] = 'comment'
                ecomment.update(self.get_grimoire_fields(created, ecomment['type']))
                ecomment.pop('is_meetup_meetup')
                # event host fields: author of the event
                member = comment['member']
                if 'photo' in member:
                    ecomment['member_photo_url'] = member['photo']['photo_link']
                    ecomment['member_photo_id'] = member['photo']['id']
                    ecomment['member_photo_type'] = member['photo']['type']
                if 'event_context' in member:
                    ecomment['member_is_host'] = member['event_context']['host']
                ecomment['member_id'] = member['id']
                ecomment['member_name'] = member['name']
                ecomment['member_url'] = "https://www.meetup.com/members/" + str(member['id'])
                comments_enrich.append(ecomment)

                if self.sortinghat:
                    ecomment.update(self.get_item_sh(comment))

        return comments_enrich

    def get_field_unique_id_comment(self):
        return "id"

    def get_rich_item_rsvps(self, item):
        rsvps_enrich = []
        if 'rsvps' in item['data']:
            for rsvp in item['data']['rsvps']:
                ersvp = self.get_rich_item(item)  # reuse all fields from item
                ersvp['type'] = 'rsvp'
                created = unixtime_to_datetime(rsvp['created']/1000).isoformat()
                ersvp.update(self.get_grimoire_fields(created, ersvp['type']))
                ersvp.pop('is_meetup_meetup')
                # event host fields: author of the event
                member = rsvp['member']
                if 'photo' in member:
                    ersvp['member_photo_url'] = member['photo']['photo_link']
                    ersvp['member_photo_id'] = member['photo']['id']
                    ersvp['member_photo_type'] = member['photo']['type']
                ersvp['member_is_host'] = member['event_context']['host']
                ersvp['member_id'] = member['id']
                ersvp['member_name'] = member['name']
                ersvp['member_url'] = "https://www.meetup.com/members/" + str(member['id'])

                ersvp['id'] = ersvp['id'] + "_" + str(member['id'])
                ersvp['url'] = "https://www.meetup.com/members/" + str(member['id'])

                ersvp['rsvps_guests'] = rsvp['guests']
                ersvp['rsvps_updated'] = rsvp['updated']
                ersvp['rsvps_response']	= rsvp['response']

                if self.sortinghat:
                    ersvp.update(self.get_item_sh(rsvp))

                rsvps_enrich.append(ersvp)

        return rsvps_enrich

    def get_field_unique_id_rsvps(self):
        return "id"

    def enrich_items(self, items):
        super(MeetupEnrich, self).enrich_items(items)

        # And now for each item we want also the rsvps and comments items
        ncom = 0
        nrsvps = 0
        rich_item_comments = []
        rich_item_rsvps = []

        for item in items:
            rich_item_comments += self.get_rich_item_comments(item)
            rich_item_rsvps += self.get_rich_item_rsvps(item)

        if rich_item_comments:
            ncom += self.elastic.bulk_upload(rich_item_comments,
                                             self.get_field_unique_id_comment())
        if rich_item_rsvps:
            nrsvps += self.elastic.bulk_upload(rich_item_rsvps,
                                               self.get_field_unique_id_rsvps())

        logging.info("Total comments enriched: %i", ncom)
        logging.info("Total nrsvps enriched: %i", nrsvps)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Ocean lib
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

"""Ocean feeder for Elastic from  Perseval data"""


import inspect
import json
import logging
import requests

from datetime import datetime
from ..elk.utils import unixtime_to_datetime


class ElasticOcean(object):

    @classmethod
    def add_params(cls, cmdline_parser):
        """ Shared params in all backends """

        parser = cmdline_parser

        parser.add_argument("-e", "--elastic_url",  default="http://127.0.0.1:9200",
                            help="Host with elastic search" +
                            "(default: http://127.0.0.1:9200)")
        parser.add_argument("--elastic_url-enrich",
                            help="Host with elastic search and enriched indexes")

    def __init__(self, perceval_backend, from_date=None, fetch_cache=False,
                 project=None, insecure=True, offset=None):

        self.perceval_backend = perceval_backend
        self.last_update = None  # Last update in ocean items index for feed
        self.from_date = from_date  # fetch from_date
        self.offset = offset  # fetch from offset
        self.fetch_cache = fetch_cache  # fetch from cache
        self.project = project  # project to be used for this data source
        self.filter_raw = None  # to filter raw items from Ocean

        self.requests = requests.Session()
        if insecure:
            requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)
            self.requests.verify = False


    def set_elastic(self, elastic):
        """ Elastic used to store last data source state """
        self.elastic = elastic

    def get_field_date(self):
        """ Field with the update in the JSON items. Now the same in all. """
        return "metadata__updated_on"

    def get_field_unique_id(self):
        return "ocean-unique-id"

    def get_elastic_mappings(self):
        """ specific mappings implemented in each data source """
        mapping = '{}'

        return {"items":mapping}

    def get_elastic_analyzers(self):
        """ Custom analyzers for our indexes  """

        return None

    def get_last_update_from_es(self, _filter = None):
        last_update = self.elastic.get_last_date(self.get_field_date(), _filter)

        return last_update

    def get_connector_name(self):
        """ Find the name for the current connector """
        from ..utils import get_connector_name
        return get_connector_name(type(self))

    @classmethod
    def get_p2o_params_from_url(cls, url):
        """ Get the p2o params given a URL for the data source """
        return {"url": url}

    @classmethod
    def get_perceval_params_from_url(cls, url):
        """ Get the perceval params given a URL for the data source """
        return [url]

    @classmethod
    def get_arthur_params_from_url(cls, url):
        """ Get the arthur params given a URL for the data source """
        return {"uri": url}

    def drop_item(self, item):
        """ Drop items not to be inserted in Elastic """
        return False

    def _fix_item(self, item):
        """ Some buggy data sources need fixing (like mbox and message-id) """
        pass

    def add_update_date(self, item):
        """ All item['updated_on'] from perceval is epoch """
        updated = unixtime_to_datetime(item['updated_on'])
        timestamp = unixtime_to_datetime(item['timestamp'])
        item['metadata__updated_on'] = updated.isoformat()
        # Also add timestamp used in incremental enrichment
        item['metadata__timestamp'] = timestamp.isoformat()

    def feed(self, from_date=None, from_offset=None, category=None):
        """ Feed data in Elastic from Perceval """

        if from_date and from_offset:
            raise RuntimeError("Can't not feed using from_date and from_offset.")

        # Check if backend supports from_date
        signature = inspect.signature(self.perceval_backend.fetch)

        last_update = None
        # Filter by tag to support multi tag indexes
        filter_ = {"name": "tag",
                   "value": self.perceval_backend.tag}

        if 'from_date' in signature.parameters:
            if from_date:
                last_update = from_date
            else:
                self.last_update = self.get_last_update_from_es([filter_])
                last_update = self.last_update

            logging.info("Incremental from: %s", last_update)

        offset = None
        if 'offset' in signature.parameters:
            if from_offset:
                offset = from_offset
            else:
                offset = self.elastic.get_last_offset("offset", [filter_])

            if offset:
                logging.info("Incremental from: %i offset", offset)
            else:
                logging.info("Not incremental")

        task_init = datetime.now()

        items_pack = []  # to feed item in packs
        drop = 0
        added = 0
        if self.fetch_cache:
            items = self.perceval_backend.fetch_from_cache()
        else:
            if last_update:
                # if offset used for incremental do not use date
                # Perceval backend from_date must not include timezone
                # It always uses the server datetime
                last_update = last_update.replace(tzinfo=None)
                if category:
                    items = self.perceval_backend.fetch(from_date=last_update, category=category)
                else:
                    items = self.perceval_backend.fetch(from_date=last_update)
            elif offset:
                if category:
                    items = self.perceval_backend.fetch(offset=offset, category=category)
                else:
                    items = self.perceval_backend.fetch(offset=offset)
            else:
                if category:
                    items = self.perceval_backend.fetch(category=category)
                else:
                    items = self.perceval_backend.fetch()

        for item in items:
            # print("%s %s" % (item['url'], item['lastUpdated_date']))
            # Add date field for incremental analysis if needed
            self.add_update_date(item)
            self._fix_item(item)
            if self.project:
                item['project'] = self.project
            if len(items_pack) >= self.elastic.max_items_bulk:
                self._items_to_es(items_pack)
                items_pack = []
            if not self.drop_item(item):
                items_pack.append(item)
                added += 1
            else:
                drop +=1
        self._items_to_es(items_pack)


        total_time_min = (datetime.now()-task_init).total_seconds()/60

        logging.debug("Added %i items to ocean", added)
        logging.debug("Dropped %i items using drop_item filter" % (drop))
        logging.info("Finished in %.2f min" % (total_time_min))

        return self


    def _items_to_es(self, json_items):
        """ Append items JSON to ES (data source state) """

        if len(json_items) == 0:
            return

        logging.info("Adding items to Ocean for %s (%i items)" %
                      (self, len(json_items)))

        field_id = self.get_field_unique_id()

        self.elastic.bulk_upload_sync(json_items, field_id)

    # Iterator
    def _get_elastic_items(self):
        """ Get the items from the index related to the backend """

        url = self.elastic.index_url
        # 1 minute to process the results of size items
        # In gerrit enrich with 500 items per page we need >1 min
        # In Mozilla ES in Amazon we need 10m
        max_process_items_pack_time = "10m"  # 10 minutes
        url += "/_search?scroll=%s&size=%i" % (max_process_items_pack_time,
                                               self.elastic_page)

        if self.elastic_scroll_id:
            """ Just continue with the scrolling """
            url = self.elastic.url
            url += "/_search/scroll"
            scroll_data = {
                "scroll" : max_process_items_pack_time,
                "scroll_id" : self.elastic_scroll_id
                }
            r = self.requests.post(url, data=json.dumps(scroll_data))
        else:
            filters = "{}"
            # If using a perceval backends always filter by tag to support multi tag
            if self.perceval_backend and self.perceval_backend.tag:
                filters = '''
                    {"term":
                        { "tag" : "%s"  }
                    }
                ''' % (self.perceval_backend.tag)

            if self.filter_raw:
                filters += '''
                    , {"term":
                        { "%s":"%s"  }
                    }
                ''' % (self.filter_raw['name'], self.filter_raw['value'])

            if self.from_date:
                date_field = self.get_field_date()
                from_date = self.from_date.isoformat()

                filters += '''
                    , {"range":
                        {"%s": {"gte": "%s"}}
                    }
                ''' % (date_field, from_date)
            elif self.offset:
                filters += '''
                    , {"range":
                        {"offset": {"gte": %i}}
                    }
                ''' % (self.offset)


            # order_field = 'metadata__updated_on'
            order_field = 'metadata__timestamp'  # to avoid holes
            order_query = ''
            if self.perceval_backend:
                # logstash backends does not have the order_field
                order_query = ', "sort": { "%s": { "order": "asc" }} ' % order_field

            query = """
            {
                "query": {
                    "bool": {
                        "must": [%s]
                    }
                } %s
            }
            """ % (filters, order_query)

            logging.debug("%s %s", url, query)

            r = self.requests.post(url, data=query)

        items = []
        try:
            rjson = r.json()
        except:
            logging.error("No JSON found in %s" % (r.text))
            logging.error("No results found from %s" % (url))

        if rjson and "_scroll_id" in rjson:
            self.elastic_scroll_id = rjson["_scroll_id"]
        else:
            self.elastic_scroll_id = None

        if rjson and "hits" in rjson:
            for hit in rjson["hits"]["hits"]:
                items.append(hit['_source'])
        else:
            logging.error("No results found from %s" % (url))

        return items

    def set_filter_raw(self, filter_raw):
        """ Filter to be used when getting items from Ocean index """

        self.filter_raw = filter_raw


    def __iter__(self):

        self.elastic_scroll_id = None
        # In large projects like Eclipse commits, 100 is too much
        # self.elastic_page = 100
        self.elastic_page = 10
        self.iter_items = self._get_elastic_items()

        return self

    def __next__(self):

        if len(self.iter_items) > 0:
            return self.iter_items.pop()
        else:
            if self.elastic_scroll_id:
                self.iter_items = self._get_elastic_items()
            if len(self.iter_items) > 0:
                return self.__next__()
            else:
                raise StopIteration



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015-2016 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from datetime import datetime

from dateutil import parser

from .enrich import Enrich, metadata

from .utils import get_time_diff_days, unixtime_to_datetime


TASK_OPEN_STATUS = 'open'
TASK_CLOSED_STATUS = 'resolved'

class PhabricatorEnrich(Enrich):

    roles = ['authorData', 'ownerData']

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                     db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map, db_user, db_password, db_host)

        self.phab_ids_names = {}  # To convert from phab ids to phab names

    def get_field_event_unique_id(self):
        return "transactionID"

    def get_field_author(self):
        return "authorData"

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "main_description_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                },
                "author_roles_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                },
                "assigned_to_roles_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                 },
                "author_roles_analyzed": {
                   "type": "string",
                   "index":"analyzed"
                 },
                "tags_analyzed": {
                   "type": "string",
                   "index":"analyzed"
                 },
                "tags_custom_analyzed" : {
                    "type" : "string",
                    "analyzer" : "comma"
                }
           }
        } """

        return {"items":mapping}

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        if 'authorData' in item['data']['fields']:
            user = self.get_sh_identity(item['data']['fields']['authorData'])
            identities.append(user)

        if 'ownerData' in item['data']['fields']:
            user = self.get_sh_identity(item['data']['fields']['ownerData'])
            identities.append(user)

        return identities

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        user = item
        if 'data' in item and type(item) == dict:
            user = item['data']['fields'][identity_field]

        identity['email'] = None
        identity['username'] = user['userName']
        identity['name'] = user['realName']

        return identity

    def get_users_data(self, item):
        """ If user fields are inside the global item dict """
        if 'data' in item:
            users_data = item['data']['fields']
        else:
            # the item is directly the data (kitsune answer)
            users_data = item
        return users_data

    def get_rich_events(self, item):
        """
        In the events there are some common fields with the task. The name
        of the field must be the same in the task and in the event
        so we can filer using it in task and event at the same time.

        * Fields that don't change: the field does not change with the events
        in a task so the value is always the same in the events of a task.

        * Fields that change: the value of teh field changes with events
        """
        events = []

        # To get values from the task
        eitem = self.get_rich_item(item)

        # Fields that don't change never
        task_fields_nochange = ['author_userName', 'creation_date', 'url', 'id', 'bug_id']

        # Follow changes in this fields
        task_fields_change = ['priority_value', 'status', 'assigned_to_userName', 'tags_custom_analyzed']
        task_change = {}
        for f in task_fields_change:
            task_change[f] = None
        task_change['status'] = TASK_OPEN_STATUS
        task_change['tags_custom_analyzed'] = eitem['tags_custom_analyzed']

        # Events are in transactions field (changes in fields)
        transactions = item['data']['transactions']
        for t in transactions:
            event = {}
            # Needed for incremental updates from the item
            event['metadata__updated_on'] = item['metadata__updated_on']
            event['origin'] = item['origin']
            # Real event data
            event['transactionID'] = t['transactionID']
            event['type'] = t['transactionType']
            event['username'] = None
            if 'authorData' in t and 'userName' in t['authorData']:
                event['event_author_name'] = t['authorData']['userName']
            event['update_date'] = unixtime_to_datetime(float(t['dateCreated'])).isoformat()
            event['oldValue'] = ''
            event['newValue'] = ''
            if event['type'] == 'core:edge':
                for val in t['oldValue']:
                    if val in self.phab_ids_names:
                        val = self.phab_ids_names[val]
                    event['oldValue'] += "," + val
                event['oldValue'] = event['oldValue'][1:]  # remove first comma
                for val in t['newValue']:
                    if val in self.phab_ids_names:
                        val = self.phab_ids_names[val]
                    event['newValue'] += "," + val
                event['newValue'] = event['newValue'][1:]  # remove first comma
            elif event['type'] in  ['status', 'description', 'priority', 'reassign', 'title', 'space', 'core:create', 'parent']:
                # Convert to str so the field is always a string
                event['oldValue'] = str(t['oldValue'])
                if event['oldValue'] in self.phab_ids_names:
                    event['oldValue'] = self.phab_ids_names[event['oldValue']]
                event['newValue'] = str(t['newValue'])
                if event['newValue'] in self.phab_ids_names:
                    event['newValue'] = self.phab_ids_names[event['newValue']]
            elif event['type'] == 'core:comment':
                event['newValue'] = t['comments']
            elif event['type'] == 'core:subscribers':
                event['newValue']= ",".join(t['newValue'])
            else:
                # logging.debug("Event type %s old to new value not supported", t['transactionType'])
                pass

            for f in task_fields_nochange:
                # The field name must be the same than in task for filtering
                event[f] = eitem[f]

            # To track history of some fields
            if event['type'] in ['status']:
                task_change['status'] = event['newValue']
            elif event['type'] == 'priority':
                task_change['priority'] =  event['newValue']
            elif event['type'] == 'core:edge':
                task_change['tags_custom_analyzed'] =  event['newValue']
            if event['type'] in  ['reassign']:
                # Try to get the userName and not the user id
                if event['newValue'] in self.phab_ids_names:
                    task_change['assigned_to_userName'] = self.phab_ids_names[event['newValue']]
                    event['newValue'] = task_change['assigned_to_userName']
                else:
                    task_change['assigned_to_userName'] = event['newValue']
                if event['oldValue'] in self.phab_ids_names:
                    # Try to get the userName and not the user id
                    event['oldValue'] = self.phab_ids_names[event['oldValue']]


            for f in task_change:
                event[f] = task_change[f]

            events.append(event)

        return events


    def __fill_phab_ids(self, item):
        """ Get mappings between phab ids and names """
        for p in item['projects']:
            self.phab_ids_names[p['phid']] = p['name']
        self.phab_ids_names[item['fields']['authorData']['phid']] = item['fields']['authorData']['userName']
        if 'ownerData' in item['fields']:
            self.phab_ids_names[item['fields']['ownerData']['phid']] = item['fields']['ownerData']['userName']
        if 'priority' in item['fields']:
            val = item['fields']['priority']['value']
            self.phab_ids_names[str(val)] = item['fields']['priority']['name']
        for t in item['transactions']:
            if 'userName' in t['authorData']:
                self.phab_ids_names[t['authorData']['phid']] = t['authorData']['userName']
            elif 'name' in t['authorData']:
                # Herald
                self.phab_ids_names[t['authorData']['phid']] = t['authorData']['name']

    def get_rich_item(self, item):
        eitem = {}

        self.__fill_phab_ids(item['data'])

        # metadata fields to copy
        copy_fields = ["metadata__updated_on", "metadata__timestamp", "ocean-unique-id", "origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None
        # The real data
        phab_item = item['data']

        # data fields to copy
        copy_fields = ["phid", "id", "type"]
        for f in copy_fields:
            if f in phab_item:
                eitem[f] = phab_item[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {
            "id": "bug_id"
        }
        for f in map_fields:
            if f in phab_item:
                eitem[map_fields[f]] = phab_item[f]
            else:
                eitem[map_fields[f]] = None

        eitem['num_changes'] = len(phab_item['transactions'])

        if 'authorData' in phab_item['fields']:
            eitem['author_roles'] = ",".join(phab_item['fields']['authorData']['roles'])
            eitem['author_roles_analyzed'] = eitem['author_roles']
            eitem['author_userName'] = phab_item['fields']['authorData']['userName']
            eitem['author_realName'] = phab_item['fields']['authorData']['realName']
        if 'ownerData' in phab_item['fields']:
            eitem['assigned_to_roles'] = ",".join(phab_item['fields']['ownerData']['roles'])
            eitem['assigned_to_roles_analyzed'] = eitem['assigned_to_roles']
            eitem['assigned_to_userName'] = phab_item['fields']['ownerData']['userName']
            eitem['assigned_to_realName'] = phab_item['fields']['ownerData']['realName']

        eitem['priority'] = phab_item['fields']['priority'] ['name']
        eitem['priority_value'] = phab_item['fields']['priority']['value']
        eitem['status'] = phab_item['fields']['status']['value']
        eitem['creation_date'] = unixtime_to_datetime(phab_item['fields']['dateCreated']).isoformat()
        eitem['modification_date'] = unixtime_to_datetime(phab_item['fields']['dateModified']).isoformat()
        eitem['update_date'] = unixtime_to_datetime(item['updated_on']).isoformat()
        # raise
        eitem['main_description'] = phab_item['fields']['name']
        eitem['main_description_analyzed'] = eitem['main_description']
        eitem['url'] = eitem['origin']+"/T"+str(eitem['bug_id'])

        # Time to assign (time to open -> time to assign)
        eitem['time_to_assign_days'] = None
        # Time to attend (time to assign-> time to first activity from assignee)
        eitem['time_to_attend_days'] = None
        # Time to close (time open -> time last updated for closed tasks)
        # We can improve it later using events: time open event -> time resolved event
        eitem['time_to_close_days'] = None
        if eitem['status'] not in [TASK_OPEN_STATUS, 'spite', 'stalled']:
            eitem['time_to_close_days'] = \
                get_time_diff_days(eitem['creation_date'], eitem['update_date'])
        # Time open (time to open -> now): with painless
        # Time open using the enrich date. Field needed for filtering.
        eitem['time_open_days_enrich'] = get_time_diff_days(eitem['creation_date'], datetime.utcnow())
        # Time from last update (time last update -> now): with painless

        eitem['changes'] = len(phab_item['transactions'])
        # Number of assignments changes
        eitem['changes_assignment'] = 0
        # Number of assignees in the changes
        eitem['changes_assignee_number'] = 0
        # List the changes assignees
        changes_assignee_list = []
        first_assignee_phid = None
        first_assignee_date = None
        # We need to revert them to go from older to newer
        phab_item['transactions'].reverse()
        for change in phab_item['transactions']:
            change_date = unixtime_to_datetime(float(change['dateCreated'])).isoformat()
            if change["transactionType"] == "reassign":
                if not eitem['time_to_assign_days']:
                    eitem['time_to_assign_days'] = get_time_diff_days(eitem['creation_date'], change_date)
                    first_assignee_phid = change['newValue']
                    first_assignee_date = change_date
                if 'userName' in change['authorData'] and \
                    change['authorData']['userName'] not in changes_assignee_list:
                    changes_assignee_list.append(change['authorData']['userName'])
                eitem['changes_assignment'] += 1
            if not eitem['time_to_attend_days'] and first_assignee_phid:
                if change['authorData']['phid'] == first_assignee_phid:
                    eitem['time_to_attend_days'] = get_time_diff_days(first_assignee_date, change_date)
        eitem['changes_assignee_number'] = len(changes_assignee_list)
        eitem['changes_assignee_list'] = ','.join(changes_assignee_list)
        eitem['comments'] = 0
        for tr in phab_item['transactions']:
            if tr ['comments']:
                eitem['comments'] += 1

        eitem['tags'] = None
        for project in phab_item['projects']:
            if not eitem['tags']:
                eitem['tags'] = project['name']
            else:
                eitem['tags'] += ',' + project['name']
        eitem['tags_analyzed'] = eitem['tags']
        eitem['tags_custom_analyzed'] = eitem['tags']

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, self.roles))

        eitem.update(self.get_grimoire_fields(eitem['creation_date'], "task"))

        return eitem



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Elastic Search basic utils
#
# Copyright (C) 2016 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from datetime import datetime
from dateutil import parser
import json
import logging
import requests

from time import time, sleep

from requests.packages.urllib3.exceptions import InsecureRequestWarning

from .utils import unixtime_to_datetime

WAIT_INDEX_CREATION = 2  # number of seconds to wait for index creation

class ElasticConnectException(Exception):
    message = "Can't connect to ElasticSearch"

class ElasticWriteException(Exception):
    message = "Can't write to ElasticSearch"

class ElasticSearch(object):

    @classmethod
    def safe_index(cls, unique_id):
        """ Return a valid elastic index generated from unique_id """
        return unique_id.replace("/","_").lower()

    def __init__(self, url, index, mappings = None, clean = False,
                 insecure=True, analyzers=None):
        ''' clean: remove already existing index
            insecure: support https with invalid certificates
        '''

        self.url = url
        # Valid index for elastic
        self.index = self.safe_index(index)
        self.index_url = self.url+"/"+self.index
        self.max_items_bulk = 100
        self.wait_bulk_seconds = 2  # time to wait to complete a bulk operation

        self.requests = requests.Session()
        if insecure:
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
            self.requests.verify = False

        try:
            r = self.requests.get(self.index_url)
        except requests.exceptions.ConnectionError as ex:
            logging.error(ex)
            raise ElasticConnectException()

        if r.status_code != 200:
            # Index does no exists
            r = self.requests.put(self.index_url, data=analyzers)
            if r.status_code != 200:
                logging.error("Can't create index %s (%s)",
                              self.index_url, r.status_code)
                raise ElasticWriteException()
            else:
                logging.info("Created index " + self.index_url)
        else:
            if clean:
                self.requests.delete(self.index_url)
                self.requests.put(self.index_url, data=analyzers)
                logging.info("Deleted and created index " + self.index_url)
        if mappings:
            self.create_mappings(mappings)

    def _safe_put_bulk(self, url, bulk_json):
        """ Bulk PUT controlling unicode issues """

        try:
            self.requests.put(url, data=bulk_json)
        except UnicodeEncodeError:
            # Related to body.encode('iso-8859-1'). mbox data
            logging.error("Encondig error ... converting bulk to iso-8859-1")
            bulk_json = bulk_json.encode('iso-8859-1','ignore')
            self.requests.put(url, data=bulk_json)

    def bulk_upload(self, items, field_id):
        ''' Upload in controlled packs items to ES using bulk API '''

        max_items = self.max_items_bulk
        current = 0
        new_items = 0  # total items added with bulk
        bulk_json = ""

        url = self.index_url+'/items/_bulk'

        logging.debug("Adding items to %s (in %i packs)" % (url, max_items))

        for item in items:
            if current >= max_items:
                task_init = time()
                self._safe_put_bulk(url, bulk_json)
                bulk_json = ""
                new_items += current
                current = 0
                logging.debug("bulk packet sent (%.2f sec, %i total)"
                              % (time()-task_init, new_items))
            data_json = json.dumps(item)
            bulk_json += '{"index" : {"_id" : "%s" } }\n' % (item[field_id])
            bulk_json += data_json +"\n"  # Bulk document
            current += 1
        task_init = time()
        self._safe_put_bulk(url, bulk_json)
        new_items += current
        logging.debug("bulk packet sent (%.2f sec prev, %i total)"
                      % (time()-task_init, new_items))

        return new_items

    def bulk_upload_sync(self, items, field_id, sync=True):
        ''' Upload in controlled packs items to ES using bulk API
            and wait until the items appears in searches '''

        # After a bulk upload the searches are not refreshed real time
        # This method waits until the upload is visible in searches

        r = self.requests.get(self.index_url+'/_search?size=1')
        total = r.json()['hits']['total']  # Already existing items
        new_items = self.bulk_upload(items, field_id)
        if not sync:
            return
        total += new_items

        # Wait until in searches all items are returned
        # In incremental update, some items are updates not additions so
        # this check will fail. Exist ok in the timeout for this case.

        total_search = 0  # total items found with search
        search_start = datetime.now()
        while total_search != total:
            sleep(0.1)
            r = self.requests.get(self.index_url+'/_search?size=1')
            total_search = r.json()['hits']['total']
            if (datetime.now()-search_start).total_seconds() > self.wait_bulk_seconds:
                logging.debug("Bulk data does not appear as NEW after %is" % (self.wait_bulk_seconds))
                logging.debug("Probably %i item updates" % (total-total_search))
                break

    def create_mappings(self, mappings):

        for _type in mappings:

            url_map = self.index_url + "/"+_type+"/_mapping"

            # First create the manual mappings
            if mappings[_type] != '{}':
                r = self.requests.put(url_map, data=mappings[_type])
                if r.status_code != 200:
                    logging.error("Error creating ES mappings %s", r.text)

            # By default all strings are not analyzed
            not_analyze_strings = """
            {
              "dynamic_templates": [
                { "notanalyzed": {
                      "match": "*",
                      "match_mapping_type": "string",
                      "mapping": {
                          "type":        "string",
                          "index":       "not_analyzed"
                      }
                   }
                }
              ]
            } """
            r = self.requests.put(url_map, data=not_analyze_strings)

    def get_last_date(self, field, _filters = []):
        '''
            :field: field with the data
            :_filter: additional filter to find the date
        '''

        last_date = self.get_last_item_field(field, _filters=_filters)

        return last_date

    def get_last_offset(self, field, _filters = []):
        '''
            :field: field with the data
            :_filter: additional filter to find the date
        '''

        offset = self.get_last_item_field(field, _filters=_filters, offset=True)

        return offset

    def get_last_item_field(self, field, _filters = [], offset = False):
        '''
            :field: field with the data
            :_filters: additional filters to find the date
            :offset: Return offset field insted of date field
        '''

        last_value = None

        url = self.index_url
        url += "/_search"

        data_query = ''
        for _filter in _filters:
            if not _filter:
                continue
            data_query += '''
                "query" : {
                    "term" : { "%s" : "%s"  }
                 },
            ''' % (_filter['name'], _filter['value'])

        data_agg = '''
            "aggs": {
                "1": {
                  "max": {
                    "field": "%s"
                  }
                }
            }
        ''' % (field)

        data_json = '''
        { "size": 0, %s  %s
        } ''' % (data_query, data_agg)

        logging.debug("%s %s", url, data_json)
        res = self.requests.post(url, data=data_json)
        res_json = res.json()

        if 'aggregations' in res_json:
            last_value = res_json["aggregations"]["1"]["value"]

            if not offset:
                if "value_as_string" in res_json["aggregations"]["1"]:
                    last_value = res_json["aggregations"]["1"]["value_as_string"]
                    last_value = parser.parse(last_value)
                else:
                    last_value = res_json["aggregations"]["1"]["value"]
                    if last_value:
                        try:
                            last_value = unixtime_to_datetime(last_value)
                        except ValueError:
                            # last_value is in microsecs
                            last_value = unixtime_to_datetime(last_value/1000)
        return last_value



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging
import re
import time

import requests

from dateutil import parser

from .enrich import Enrich, metadata

try:
    from .sortinghat import SortingHat
    SORTINGHAT_LIBS = True
except ImportError:
    SORTINGHAT_LIBS = False

GITHUB = 'https://github.com/'
SH_GIT_COMMIT = 'github-commit'
DEMOGRAPHY_COMMIT_MIN_DATE='1980-01-01'

class GitEnrich(Enrich):

    roles = ['Author', 'Commit']

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)

        self.studies = [self.enrich_demography]

        # GitHub API management
        self.github_token = None
        self.github_logins = {}
        self.github_logins_committer_not_found = 0
        self.github_logins_author_not_found = 0
        self.rate_limit = None
        self.rate_limit_reset_ts = None
        self.min_rate_to_sleep = 100  # if pending rate < 100 sleep

    def set_github_token(self, token):
        self.github_token = token

    def get_field_author(self):
        return "Author"

    def get_field_unique_id(self):
        return "ocean-unique-id"

    def get_fields_uuid(self):
        return ["author_uuid", "committer_uuid"]

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
               "message_analyzed": {
                  "type": "string",
                  "index":"analyzed"
               }
           }
       }"""

        return {"items":mapping}

    def get_identities(self, item):
        """ Return the identities from an item.
            If the repo is in GitHub, get the usernames from GitHub. """
        identities = []

        def add_sh_github_identity(user, user_field, rol):
            """ Add a new github identity to SH if it does not exists """
            github_repo = None
            if GITHUB in item['origin']:
                github_repo = item['origin'].replace(GITHUB,'')
                github_repo = re.sub('.git$', '', github_repo)
            if not github_repo:
                return

            # Try to get the identity from SH
            user_data = item['data'][user_field]
            sh_identity = SortingHat.get_github_commit_username(self.sh_db, user, SH_GIT_COMMIT)
            if not sh_identity:
                # Get the usename from GitHub
                gh_username = self.get_github_login(user_data, rol, commit_hash, github_repo)
                # Create a new SH identity with name, email from git and username from github
                logging.debug("Adding new identity %s to SH %s: %s", gh_username, SH_GIT_COMMIT, user)
                user = self.get_sh_identity(user_data)
                user['username'] = gh_username
                SortingHat.add_identity(self.sh_db, user, SH_GIT_COMMIT)
            else:
                if user_data not in self.github_logins:
                    self.github_logins[user_data] = sh_identity['username']
                    logging.debug("GitHub-commit exists. username:%s user:%s",
                                  sh_identity['username'], user_data)

        commit_hash = item['data']['commit']

        if item['data']['Author']:
            user = self.get_sh_identity(item['data']["Author"])
            identities.append(user)
            if self.github_token:
                add_sh_github_identity(user, 'Author', 'author')
        if item['data']['Commit']:
            user = self.get_sh_identity(item['data']['Commit'])
            identities.append(user)
            if self.github_token:
                add_sh_github_identity(user, 'Commit', 'committer')

        return identities

    def get_sh_identity(self, item, identity_field=None):
        # John Smith <john.smith@bitergia.com>
        identity = {}

        git_user = item  # by default a specific user dict is expected
        if 'data' in item and type(item) == dict:
            git_user = item['data'][identity_field]

        name = git_user.split("<")[0]
        name = name.strip()  # Remove space between user and email
        email = git_user.split("<")[1][:-1]
        identity['username'] = None
        identity['email'] = email
        identity['name'] = name

        return identity

    def get_project_repository(self, eitem):
        return eitem['origin']

    def get_github_login(self, user, rol, commit_hash, repo):
        """ rol: author or committer """
        login = None
        try:
            login = self.github_logins[user]
        except KeyError:
            # Get the login from github API
            GITHUB_API_URL = "https://api.github.com"
            commit_url = GITHUB_API_URL+"/repos/%s/commits/%s" % (repo, commit_hash)
            headers = {'Authorization': 'token ' + self.github_token}

            r = self.requests.get(commit_url, headers=headers)

            self.rate_limit = int(r.headers['X-RateLimit-Remaining'])
            self.rate_limit_reset_ts = int(r.headers['X-RateLimit-Reset'])
            logging.debug("Rate limit pending: %s", self.rate_limit)
            if self.rate_limit <= self.min_rate_to_sleep:
                seconds_to_reset = self.rate_limit_reset_ts - int(time.time()) + 1
                if seconds_to_reset < 0:
                    seconds_to_reset = 0
                cause = "GitHub rate limit exhausted."
                logging.info("%s Waiting %i secs for rate limit reset.", cause, seconds_to_reset)
                time.sleep(seconds_to_reset)
                # Retry once we have rate limit
                r = self.requests.get(commit_url, headers=headers)

            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as ex:
                # commit not found probably or rate limit exhausted
                logging.error("Can't find commit %s %s", commit_url, ex)
                return login

            commit_json = r.json()
            author_login = None
            if 'author' in commit_json and commit_json['author']:
                author_login = commit_json['author']['login']
            else:
                self.github_logins_author_not_found += 1

            user_login = None
            if 'committer' in commit_json and commit_json['committer']:
                user_login = commit_json['committer']['login']
            else:
                self.github_logins_committer_not_found += 1

            if rol == "author":
                login = author_login
            elif rol == "committer":
                login = user_login
            else:
                logging.error("Wrong rol: %s" % (rol))
                raise RuntimeError

            self.github_logins[user] = login
            logging.debug("%s is %s in github (not found %i authors %i committers )", user, login,
                          self.github_logins_author_not_found,
                          self.github_logins_committer_not_found)

        return login

    @metadata
    def get_rich_item(self, item):
        eitem = {}
        # metadata fields to copy
        copy_fields = ["metadata__updated_on","metadata__timestamp","ocean-unique-id","origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None
        # The real data
        commit = item['data']
        # data fields to copy
        copy_fields = ["message","Author"]
        for f in copy_fields:
            if f in commit:
                eitem[f] = commit[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"commit": "hash","message":"message_analyzed","Commit":"Committer"}
        for fn in map_fields:
            if fn in commit:
                eitem[map_fields[fn]] = commit[fn]
            else:
                eitem[map_fields[fn]] = None
        eitem['hash_short'] = eitem['hash'][0:6]
        # Enrich dates
        author_date = parser.parse(commit["AuthorDate"])
        commit_date = parser.parse(commit["CommitDate"])
        eitem["author_date"] = author_date.replace(tzinfo=None).isoformat()
        eitem["commit_date"] = commit_date.replace(tzinfo=None).isoformat()
        eitem["utc_author"] = (author_date-author_date.utcoffset()).replace(tzinfo=None).isoformat()
        eitem["utc_commit"] = (commit_date-commit_date.utcoffset()).replace(tzinfo=None).isoformat()
        eitem["tz"]  = int(commit_date.strftime("%z")[0:3])
        # Other enrichment
        eitem["repo_name"] = item["origin"]
        # Number of files touched
        eitem["files"] = 0
        # Number of lines added and removed
        lines_added = 0
        lines_removed = 0
        for cfile in commit["files"]:
            if 'action' not in cfile:
                # merges are not counted
                continue
            eitem["files"] += 1
            if 'added' in cfile and 'removed' in cfile:
                try:
                    lines_added += int(cfile["added"])
                    lines_removed += int(cfile["removed"])
                except ValueError:
                    # logging.warning(cfile)
                    continue
        eitem["lines_added"] = lines_added
        eitem["lines_removed"] = lines_removed
        eitem["lines_changed"] = lines_added + lines_removed

        # author_name and author_domain are added always
        identity  = self.get_sh_identity(commit["Author"])
        eitem["author_name"] = identity['name']
        eitem["author_domain"] = self.get_identity_domain(identity)

        # committer data
        identity  = self.get_sh_identity(commit["Commit"])
        eitem["committer_name"] = identity['name']
        eitem["committer_domain"] = self.get_identity_domain(identity)

        # title from first line
        if 'message' in commit:
            eitem["title"] = commit['message'].split('\n')[0]
        else:
            eitem["title"] = None

        # If it is a github repo, include just the repo string
        if GITHUB in item['origin']:
            eitem['github_repo'] = item['origin'].replace(GITHUB,'')
            eitem['github_repo'] = re.sub('.git$', '', eitem['github_repo'])
            eitem["url_id"] = eitem['github_repo']+"/commit/"+eitem['hash']

        if 'project' in item:
            eitem['project'] = item['project']

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, self.roles))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(commit["AuthorDate"], "commit"))

        return eitem

    def enrich_demography(self, from_date=None):
        logging.debug("Doing demography enrich from %s", self.elastic.index_url)

        if from_date:
            # The from_date must be max author_max_date
            from_date = self.elastic.get_last_item_field("author_max_date")
            logging.debug("Demography since: %s", from_date)

        date_field = self.get_field_date()

        # Don't use commits before DEMOGRAPHY_COMMIT_MIN_DATE
        filters = '''
        {"range":
            {"%s": {"gte": "%s"}}
        }
        ''' % (date_field, DEMOGRAPHY_COMMIT_MIN_DATE)

        if from_date:
            from_date = from_date.isoformat()

            filters += '''
            ,
            {"range":
                {"%s": {"gte": "%s"}}
            }
            ''' % (date_field, from_date)

        query = """
        "query": {
            "bool": {
                "must": [%s]
            }
        },
        """ % (filters)


        # First, get the min and max commit date for all the authors
        # Limit aggregations: https://github.com/elastic/elasticsearch/issues/18838
        # 10000 seems to be a sensible number of the number of people in git
        es_query = """
        {
          %s
          "size": 0,
          "aggs": {
            "author": {
              "terms": {
                "field": "Author",
                "size": 10000
              },
              "aggs": {
                "min": {
                  "min": {
                    "field": "utc_commit"
                  }
                },
                "max": {
                  "max": {
                    "field": "utc_commit"
                  }
                }
              }
            }
          }
        }
        """ % (query)

        logging.debug(es_query)

        r = self.requests.post(self.elastic.index_url+"/_search", data=es_query, verify=False)
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as ex:
            logging.error("Error getting authors mix and max date. Demography aborted.")
            logging.error(ex)
            return

        authors = r.json()['aggregations']['author']['buckets']

        author_items = []  # items from author with new date fields added
        nauthors_done = 0
        author_query = """
        {
            "query": {
                "bool": {
                    "must": [
                        {"term":
                            { "Author" : ""  }
                        }
                        ]
                }
            }

        }
        """
        author_query_json = json.loads(author_query)


        for author in authors:
            # print("%s: %s %s" % (author['key'], author['min']['value_as_string'], author['max']['value_as_string']))
            # Time to add all the commits (items) from this author
            author_query_json['query']['bool']['must'][0]['term']['Author'] = author['key']
            author_query_str = json.dumps(author_query_json)
            r = self.requests.post(self.elastic.index_url+"/_search?size=10000", data=author_query_str, verify=False)

            if "hits" not in r.json():
                logging.error("Can't find commits for %s" % (author['key']))
                print(r.json())
                print(author_query)
                continue
            for item in r.json()["hits"]["hits"]:
                new_item = item['_source']
                new_item.update(
                    {"author_min_date":author['min']['value_as_string'],
                     "author_max_date":author['max']['value_as_string']}
                )
                author_items.append(new_item)

            if len(author_items) >= self.elastic.max_items_bulk:
                self.elastic.bulk_upload(author_items, "ocean-unique-id")
                author_items = []

            nauthors_done += 1
            logging.info("Authors processed %i/%i" % (nauthors_done, len(authors)))

        self.elastic.bulk_upload(author_items, "ocean-unique-id")

        logging.debug("Completed demography enrich from %s" % (self.elastic.index_url))



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Grimoire Arthur lib.
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import inspect
import logging
import requests
import traceback

from datetime import datetime
from dateutil import parser

from .ocean.conf import ConfOcean
from .utils import get_elastic
from .utils import get_connectors, get_connector_from_name

logger = logging.getLogger(__name__)

def feed_backend(url, clean, fetch_cache, backend_name, backend_params,
                 es_index=None, es_index_enrich=None, project=None):
    """ Feed Ocean with backend data """

    backend = None
    repo = {}    # repository data to be stored in conf
    repo['backend_name'] = backend_name
    repo['backend_params'] = backend_params

    if es_index:
        clean = False  # don't remove index, it could be shared

    if not get_connector_from_name(backend_name):
        raise RuntimeError("Unknown backend %s" % backend_name)
    connector = get_connector_from_name(backend_name)
    klass = connector[3]  # BackendCmd for the connector

    try:
        backend_cmd = klass(*backend_params)

        backend = backend_cmd.backend
        ocean_backend = connector[1](backend, fetch_cache=fetch_cache, project=project)

        logger.info("Feeding Ocean from %s (%s)", backend_name, backend.tag)

        if not es_index:
            es_index = backend_name + "_" + backend.tag
        elastic_ocean = get_elastic(url, es_index, clean, ocean_backend)

        ocean_backend.set_elastic(elastic_ocean)

        ConfOcean.set_elastic(elastic_ocean)

        repo['repo_update_start'] = datetime.now().isoformat()

        # perceval backends fetch params
        offset = None
        from_date = None
        category = None

        signature = inspect.signature(backend.fetch)

        if 'from_date' in signature.parameters:
            try:
                # Support perceval pre and post BackendCommand refactoring
                from_date = backend_cmd.from_date
            except AttributeError:
                from_date = backend_cmd.parsed_args.from_date

        if 'offset' in signature.parameters:
            try:
                offset = backend_cmd.offset
            except AttributeError:
                offset = backend_cmd.parsed_args.offset

        if 'category' in signature.parameters:
            try:
                category = backend_cmd.category
            except AttributeError:
                category = backend_cmd.parsed_args.category

        # from_date param support
        if offset and category:
            ocean_backend.feed(from_offset=offset, category=category)
        elif offset:
            ocean_backend.feed(from_offset=offset)
        elif from_date and from_date.replace(tzinfo=None) != parser.parse("1970-01-01"):
            if category:
                ocean_backend.feed(from_date, category=category)
            else:
                ocean_backend.feed(from_date)
        elif category:
            ocean_backend.feed(category=category)
        else:
            ocean_backend.feed()

    except Exception as ex:
        if backend:
            logger.error("Error feeding ocean from %s (%s): %s" %
                          (backend_name, backend.tag, ex))
            # this print makes blackbird fails
            traceback.print_exc()
        else:
            logger.error("Error feeding ocean %s" % ex)

        repo['success'] = False
        repo['error'] = str(ex)
    else:
        repo['success'] = True

    repo['repo_update'] = datetime.now().isoformat()
    repo['index'] = es_index
    repo['index_enrich'] = es_index_enrich
    repo['project'] = project

    if es_index:
        unique_id = es_index+"_"+backend.tag
        ConfOcean.add_repo(unique_id, repo)
    else:
        logger.debug("Repository not added to Ocean because errors.")
        logger.debug(backend_params)

    logger.info("Done %s " % (backend_name))


def get_items_from_uuid(uuid, enrich_backend, ocean_backend):
    """ Get all items that include uuid """

    # logging.debug("Getting items for merged uuid %s "  % (uuid))

    uuid_fields = enrich_backend.get_fields_uuid()

    terms = ""  # all terms with uuids in the enriched item

    for field in uuid_fields:
        terms += """
         {"term": {
           "%s": {
              "value": "%s"
           }
         }}
         """ % (field, uuid)
        terms += ","

    terms = terms[:-1]  # remove last , for last item

    query = """
    {"query": { "bool": { "should": [%s] }}}
    """ % (terms)

    url_search = enrich_backend.elastic.index_url+"/_search"
    url_search +="?size=1000"  # TODO get all items

    r = requests.post(url_search, data=query)

    eitems = r.json()['hits']['hits']

    if len(eitems) == 0:
        # logging.warning("No enriched items found for uuid: %s " % (uuid))
        return []

    items_ids = []

    for eitem in eitems:
        item_id = enrich_backend.get_item_id(eitem)
        # For one item several eitems could be generated
        if item_id not in items_ids:
            items_ids.append(item_id)

    # Time to get the items
    logger.debug ("Items to be renriched for merged uuids: %s" % (",".join(items_ids)))

    url_mget = ocean_backend.elastic.index_url+"/_mget"

    items_ids_query = ""

    for item_id in items_ids:
        items_ids_query += '{"_id" : "%s"}' % (item_id)
        items_ids_query += ","
    items_ids_query = items_ids_query[:-1]  # remove last , for last item

    query = '{"docs" : [%s]}' % (items_ids_query)
    r = requests.post(url_mget, data=query)

    res_items = r.json()['docs']

    items = []
    for res_item in res_items:
        if res_item['found']:
            items.append(res_item["_source"])

    return items

def refresh_projects(enrich_backend):
    logger.debug("Refreshing project field in %s", enrich_backend.elastic.index_url)
    total = 0

    eitems = enrich_backend.fetch()
    for eitem in eitems:
        new_project = enrich_backend.get_item_project(eitem)
        eitem.update(new_project)
        yield eitem
        total += 1

    logger.info("Total eitems refreshed for project field %i", total)

def refresh_identities(enrich_backend, query_string=None):
    logger.debug("Refreshing identities fields from %s", enrich_backend.elastic.index_url)
    total = 0

    for eitem in enrich_backend.fetch(query_string):
        #logger.info(eitem)
        roles = None
        try:
            roles = enrich_backend.roles
        except AttributeError:
            pass
        new_identities = enrich_backend.get_item_sh_from_id(eitem, roles)
        eitem.update(new_identities)
        yield eitem
        total += 1

    logger.info("Total eitems refreshed for identities fields %i", total)

def load_identities(ocean_backend, enrich_backend):
    try:
        from .elk.sortinghat import SortingHat
    except ImportError:
        logger.warning("SortingHat not available.")

    # First we add all new identities to SH
    items_count = 0
    new_identities = []

    for item in ocean_backend:
        items_count += 1
        # Get identities from new items to be added to SortingHat
        identities = enrich_backend.get_identities(item)
        for identity in identities:
            if identity not in new_identities:
                new_identities.append(identity)
        if items_count % 100 == 0:
            logger.debug("Processed %i items identities (%i identities)",
                          items_count, len(new_identities))
    logger.debug("TOTAL ITEMS: %i", items_count)

    logger.info("Total new identities to be checked %i", len(new_identities))

    SortingHat.add_identities(enrich_backend.sh_db, new_identities,
                              enrich_backend.get_connector_name())

    return items_count

def enrich_items(items, enrich_backend, events=False):
    total = 0

    if not events:
        total= enrich_backend.enrich_items(items)
    else:
        total = enrich_backend.enrich_events(items)
    return total

def get_last_enrich(backend_cmd, enrich_backend, filter_raw=None):
    last_enrich = None

    if backend_cmd:
        backend = backend_cmd.backend

        # Only supported in data retrieved from a perceval backend
        # Always filter by tag to support multi tag indexes

        filter_ = {"name": "tag",
                   "value": backend.tag}
        # Check if backend supports from_date
        signature = inspect.signature(backend.fetch)

        from_date = None
        if 'from_date' in signature.parameters:
            try:
                # Support perceval pre and post BackendCommand refactoring
                from_date = backend_cmd.from_date
            except AttributeError:
                from_date = backend_cmd.parsed_args.from_date

        offset = None
        if 'offset' in signature.parameters:
            try:
                offset = backend_cmd.offset
            except AttributeError:
                offset = backend_cmd.parsed_args.offset


        if from_date:
            if from_date.replace(tzinfo=None) != parser.parse("1970-01-01"):
                last_enrich = from_date
            else:
                last_enrich = enrich_backend.get_last_update_from_es([filter_, filter_raw])

        elif offset:
            if offset != 0:
                last_enrich = offset
            else:
                last_enrich = enrich_backend.get_last_offset_from_es([filter_, filter_raw])
    else:
        last_enrich = enrich_backend.get_last_update_from_es()

    return last_enrich


def get_ocean_backend(backend_cmd, enrich_backend, no_incremental, filter_raw=None):
    """ Get the ocean backend configured to start from the last enriched date """

    if no_incremental:
        last_enrich = None
    else:
        last_enrich = get_last_enrich(backend_cmd, enrich_backend, filter_raw)

    logger.debug("Last enrichment: %s", last_enrich)

    backend = None

    connector = get_connectors()[enrich_backend.get_connector_name()]

    if backend_cmd:
        backend = backend_cmd.backend
        signature = inspect.signature(backend.fetch)
        if 'from_date' in signature.parameters:
            ocean_backend = connector[1](backend, from_date=last_enrich)
        elif 'offset' in signature.parameters:
            ocean_backend = connector[1](backend, offset=last_enrich)
        else:
            ocean_backend = connector[1](backend)
    else:
        if last_enrich:
            ocean_backend = connector[1](backend, from_date=last_enrich)
        else:
            ocean_backend = connector[1](backend)

    if filter_raw:
        ocean_backend.set_filter_raw(filter_raw)

    return ocean_backend

def do_studies(enrich_backend, no_incremental=False):
    if no_incremental:
        last_enrich = None
    else:
        last_enrich = get_last_enrich(None, enrich_backend)

    try:
        for study in enrich_backend.studies:
            logger.info("Starting study: %s (from %s)", study, last_enrich)
            study(from_date=last_enrich)
    except Exception as e:
        logger.error("Problem executing study %s", study)
        traceback.print_exc()

def enrich_backend(url, clean, backend_name, backend_params, ocean_index=None,
                   ocean_index_enrich = None,
                   db_projects_map=None, json_projects_map=None,
                   db_sortinghat=None,
                   no_incremental=False, only_identities=False,
                   github_token=None, studies=False, only_studies=False,
                   url_enrich=None, events_enrich=False,
                   db_user=None, db_password=None, db_host=None,
                   do_refresh_projects=False, do_refresh_identities=False,
                   author_id=None, author_uuid=None, filter_raw=None):
    """ Enrich Ocean index """


    backend = None
    enrich_index = None

    if ocean_index or ocean_index_enrich:
        clean = False  # don't remove index, it could be shared

    if do_refresh_projects or do_refresh_identities:
        clean = False  # refresh works over the existing enriched items

    if not get_connector_from_name(backend_name):
        raise RuntimeError("Unknown backend %s" % backend_name)
    connector = get_connector_from_name(backend_name)
    klass = connector[3]  # BackendCmd for the connector

    try:
        backend = None
        backend_cmd = None
        if klass:
            # Data is retrieved from Perceval
            backend_cmd = klass(*backend_params)
            backend = backend_cmd.backend

        if ocean_index_enrich:
            enrich_index = ocean_index_enrich
        else:
            if not ocean_index:
                ocean_index = backend_name + "_" + backend.tag
            enrich_index = ocean_index+"_enrich"
        if events_enrich:
            enrich_index += "_events"

        enrich_backend = connector[2](db_sortinghat, db_projects_map, json_projects_map,
                                      db_user, db_password, db_host)
        if url_enrich:
            elastic_enrich = get_elastic(url_enrich, enrich_index, clean, enrich_backend)
        else:
            elastic_enrich = get_elastic(url, enrich_index, clean, enrich_backend)
        enrich_backend.set_elastic(elastic_enrich)
        if github_token and backend_name == "git":
            enrich_backend.set_github_token(github_token)


        # filter_raw must be converted from the string param to a dict
        filter_raw_dict = {}
        if filter_raw:
            filter_raw_dict['name'] = filter_raw.split(":")[0].replace('"','')
            filter_raw_dict['value'] = filter_raw.split(":")[1].replace('"','')

        ocean_backend = get_ocean_backend(backend_cmd, enrich_backend,
                                          no_incremental, filter_raw_dict)

        if only_studies:
            logger.info("Running only studies (no SH and no enrichment)")
            do_studies(enrich_backend, no_incremental)
        elif do_refresh_projects:
            logger.info("Refreshing project field in enriched index")
            field_id = enrich_backend.get_field_unique_id()
            eitems = refresh_projects(enrich_backend)
            enrich_backend.elastic.bulk_upload_sync(eitems, field_id)
        elif do_refresh_identities:

            if author_id:
                query_string = {}
                query_string["fields"] = 'author_id'
                query_string["query"] = author_id
            elif author_uuid:
                query_string = {}
                query_string["fields"] = 'author_uuid'
                query_string["query"] = author_uuid

            logger.info("Refreshing identities fields in enriched index")
            field_id = enrich_backend.get_field_unique_id()
            logger.info(field_id)
            eitems = refresh_identities(enrich_backend, query_string)
            enrich_backend.elastic.bulk_upload_sync(eitems, field_id)
        else:
            clean = False  # Don't remove ocean index when enrich
            elastic_ocean = get_elastic(url, ocean_index, clean, ocean_backend)
            ocean_backend.set_elastic(elastic_ocean)


            logger.info("Adding enrichment data to %s", enrich_backend.elastic.index_url)

            if db_sortinghat:
                # FIXME: This step won't be done from enrich in the future
                total_ids = load_identities(ocean_backend, enrich_backend)
                logger.info("Total identities loaded %i ", total_ids)

            if only_identities:
                logger.info("Only SH identities added. Enrich not done!")

            else:
                # Enrichment for the new items once SH update is finished
                if not events_enrich:
                    enrich_count = enrich_items(ocean_backend, enrich_backend)
                    if enrich_count:
                        logger.info("Total items enriched %i ", enrich_count)
                else:
                    enrich_count = enrich_items(ocean_backend, enrich_backend, events=True)
                    if enrich_count:
                        logger.info("Total events enriched %i ", enrich_count)
                if studies:
                    do_studies(enrich_backend)

    except Exception as ex:
        traceback.print_exc()
        if backend:
            logger.error("Error enriching ocean from %s (%s): %s",
                          backend_name, backend.tag, ex)
        else:
            logger.error("Error enriching ocean %s", ex)

    logger.info("Done %s ", backend_name)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Bugzilla Ocean feeder
#
# Copyright (C) 2016 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

'''Bugzilla Ocean feeder'''

from .elastic import ElasticOcean

class BugzillaOcean(ElasticOcean):

    def _fix_item(self, item):
        bug_id = item["data"]["bug_id"][0]['__text__']
        item["ocean-unique-id"] = bug_id+"_"+item['origin']
        # Could be used for filtering
        product = item['data']['product'][0]['__text__']
        item['product'] = product

    @classmethod
    def get_p2o_params_from_url(cls, url):
        # Bugzilla could include in the URL a filter-raw T1720
        # https://bugzilla.redhat.com/ filter-raw=product:OpenShift Origin
        params = {}

        tokens = url.split(' ', 1)  # Just split the URL not the filter
        params['url'] = tokens[0]

        if len(tokens) > 1:
            f = tokens[1].split("=")[1]
            params['filter-raw'] = f

        return params



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from dateutil import parser

from .enrich import Enrich, metadata

class TwitterEnrich(Enrich):

    def get_field_author(self):
        return  "user"

    def get_field_date(self):
        return "created_at"

    def get_field_unique_id(self):
        return "id_str"

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "hashtags_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                  },
                "text_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                  },
                  "geolocation": {
                     "type": "geo_point"
                  }
           }
        } """

        return {"items":mapping}

    def get_sh_identity(self, item, identity_field=None):
        identity = {}
        identity['username'] = None
        identity['email'] = None
        identity['name'] = None

        if identity_field in item:
            identity['username'] = item[identity_field]['screen_name']
            identity['name'] = item[identity_field]['name']
        return identity

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        user = self.get_sh_identity(item)
        identities.append(user)

        return identities

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        # The real data
        tweet = item

        # data fields to copy
        copy_fields = ["id", "id_str", "lang", "place", "retweet_count",
                       "text", "in_reply_to_user_id_str", "in_reply_to_screen_name"]
        for f in copy_fields:
            if f in tweet:
                eitem[f] = tweet[f]
            else:
                eitem[f] = None
        # Date fields
        eitem["created_at"]  = parser.parse(tweet["created_at"]).isoformat()
        # Fields which names are translated
        map_fields = {"@timestamp": "timestamp",
                      "@version": "version"
                      }
        for f in map_fields:
            if f in tweet:
                eitem[map_fields[f]] = tweet[f]
            else:
                eitem[map_fields[f]] = None

        # data fields to copy from user
        copy_fields = ["created_at", "description", "followers_count",
                       "friends_count", "id_str", "lang", "location", "name",
                       "url", "utc_offset", "verified"]
        for f in copy_fields:
            if f in tweet['user']:
                eitem["user_" + f] = tweet['user'][f]
            else:
                eitem["user_" + f] = None

        if "text" in tweet:
            eitem["text_analyzed"] = tweet["text"]

        eitem['hashtags_analyzed'] = ''
        for tag in tweet['entities']['hashtags']:
            eitem['hashtags_analyzed'] += tag['text']+","
        eitem['hashtags_analyzed'] = eitem['hashtags_analyzed'][0:-1]

        eitem['retweeted'] = 0
        if tweet['retweeted']:
            eitem['retweeted'] = 1

        eitem['url'] = "http://twitter.com/"+tweet['user']['screen_name']+"/status/"+tweet['id_str']
        eitem['user_url_twitter'] = "http://twitter.com/"+tweet['user']['screen_name']

        if self.sortinghat:
            eitem.update(self.get_item_sh(tweet))

        eitem.update(self.get_grimoire_fields(tweet["created_at"], "twitter"))

        return eitem



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging

from dateutil import parser
import email.utils

from .enrich import Enrich, metadata

class MBoxEnrich(Enrich):

    def get_field_author(self):
        return "From"

    def get_field_unique_id(self):
        return "ocean-unique-id"

    def get_fields_uuid(self):
        return ["from_uuid"]

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                 "Subject_analyzed": {
                   "type": "string",
                   "index":"analyzed"
                 }
           }
        } """

        return {"items":mapping}

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        item = item['data']
        for identity in ['From']:
            if identity in item and item[identity]:
                user = self.get_sh_identity(item[identity])
                identities.append(user)
        return identities

    def get_sh_identity(self, item, identity_field=None):
        # "From": "hwalsh at wikiledia.net (Heat Walsh)"

        identity = {}

        from_data = item
        if 'data' in item and type(item) == dict:
            from_data = item['data'][identity_field]

        # First desofuscate the email
        EMAIL_OBFUSCATION_PATTERNS = [' at ', '_at_', ' en ']
        for pattern in EMAIL_OBFUSCATION_PATTERNS:
            if from_data.find(pattern) != -1:
                from_data = from_data.replace(pattern, '@')

        from_ = email.utils.parseaddr(from_data)

        identity['username'] = None  # email does not have username
        identity['email'] = from_[1]
        identity['name'] = from_[0]
        if not identity['name']:
            identity['name'] = identity['email'].split('@')[0]
        return identity

    def get_project_repository(self, eitem):
        mls_list = eitem['origin']
        # Eclipse specific yet
        repo = "/mnt/mailman_archives/"
        repo += mls_list+".mbox/"+mls_list+".mbox"
        return repo

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        # metadata fields to copy
        copy_fields = ["metadata__updated_on","metadata__timestamp","ocean-unique-id","origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None
        # The real data
        message = item['data']

        # Fields that are the same in message and eitem
        copy_fields = ["Date","From","Subject","Message-ID"]
        for f in copy_fields:
            if f in message:
                eitem[f] = message[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"Subject": "Subject_analyzed"
                      }
        for fn in map_fields:
            if fn in message:
                eitem[map_fields[fn]] = message[fn]
            else:
                eitem[map_fields[fn]] = None

        # Enrich dates
        eitem["email_date"] = parser.parse(item["metadata__updated_on"]).isoformat()
        eitem["list"] = item["origin"]

        # Root message
        if 'In-Reply-To' in message:
            eitem["root"] = False
        else:
            eitem["root"] = True

        # Size of the message
        try:
            eitem["size"] = len(message['body']['plain'])
        except:
            eitem["size"] = None

        # Time zone
        try:
            message_date = parser.parse(message['Date'])
            eitem["tz"]  = int(message_date.strftime("%z")[0:3])
        except:
            eitem["tz"]  = None

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(message['Date'], "message"))

        return eitem

    def enrich_items(self, items):
        # Use standard method and if fails, use the old one with Unicode control
        total = 0
        try:
            total = super(MBoxEnrich, self).enrich_items(items)
        except UnicodeEncodeError:
            total = self.enrich_items_old(items)

        return total

    def enrich_items_old(self, items):
        max_items = self.elastic.max_items_bulk
        current = 0
        total = 0
        bulk_json = ""

        url = self.elastic.index_url+'/items/_bulk'

        logging.debug("Adding items to %s (in %i packs)" % (url, max_items))

        for item in items:
            if current >= max_items:
                self.requests.put(url, data=bulk_json)
                bulk_json = ""
                current = 0

            rich_item = self.get_rich_item(item)
            data_json = json.dumps(rich_item)
            bulk_json += '{"index" : {"_id" : "%s" } }\n' % \
                (rich_item[self.get_field_unique_id()])
            bulk_json += data_json +"\n"  # Bulk document
            current += 1
            total += 1
        try:
            self.requests.put(url, data = bulk_json)
        except UnicodeEncodeError:
            # Related to body.encode('iso-8859-1'). mbox data
            logging.error("Encoding error ... converting bulk to iso-8859-1")
            bulk_json = bulk_json.encode('iso-8859-1','ignore')
            self.requests.put(url, data=bulk_json)

        return total



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Bugzilla to Elastic class helper
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#


import logging

from datetime import datetime

from dateutil import parser

from .enrich import Enrich, metadata

from .utils import get_time_diff_days

class BugzillaEnrich(Enrich):

    roles = ['assigned_to', 'reporter', 'qa_contact']

    def get_field_author(self):
        return "reporter"

    def get_fields_uuid(self):
        return ["assigned_to_uuid", "reporter_uuid"]

    def get_field_unique_id(self):
        return "ocean-unique-id"

    def get_sh_identity(self, item, identity_field=None):
        """ Return a Sorting Hat identity using bugzilla user data """

        def fill_list_identity(identity, user_list_data):
            """ Fill identity with user data in first item in list """
            identity['username'] = user_list_data[0]['__text__']
            if '@' in identity['username']:
                identity['email'] = identity['username']
            if 'name' in user_list_data[0]:
                identity['name'] = user_list_data[0]['name']
            return identity

        identity = {}
        for field in ['name', 'email', 'username']:
            # Basic fields in Sorting Hat
            identity[field] = None

        user = item  # by default a specific user dict is used
        if 'data' in item and type(item) == dict:
            user = item['data'][identity_field]

        identity = fill_list_identity(identity, user)

        return identity

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        product = eitem['product']
        repo += "buglist.cgi?product="+product
        return repo

    def get_identities(self, item):
        ''' Return the identities from an item '''

        identities = []

        for rol in self.roles:
            if rol in item['data']:
                identities.append(self.get_sh_identity(item["data"][rol]))

        if 'activity' in item["data"]:
            for event in item["data"]['activity']:
                event_user = [{"__text__":event['Who']}]
                identities.append(self.get_sh_identity(event_user))
        if 'long_desc' in item["data"]:
            for comment in item["data"]['long_desc']:
                identities.append(self.get_sh_identity(comment['who']))

        return identities

    @metadata
    def get_rich_item(self, item):

        if 'bug_id' not in item['data']:
            logging.warning("Dropped bug without bug_id %s", item)
            return None

        eitem = {}

        # metadata fields to copy
        copy_fields = ["metadata__updated_on","metadata__timestamp","ocean-unique-id","origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None

        # The real data
        issue = item['data']

        # dizquierdo specification

        eitem['changes'] = len(item['data']['activity'])

        eitem['labels'] = item['data']['keywords']
        eitem['priority'] = item['data']['priority'][0]['__text__']
        eitem['severity'] = item['data']['bug_severity'][0]['__text__']
        eitem['op_sys'] = item['data']['op_sys'][0]['__text__']
        eitem['product'] = item['data']['product'][0]['__text__']
        eitem['project_name'] = item['data']['product'][0]['__text__']
        eitem['component'] = item['data']['component'][0]['__text__']
        eitem['platform'] = item['data']['rep_platform'][0]['__text__']
        if '__text__' in item['data']['resolution'][0]:
            eitem['resolution'] = item['data']['resolution'][0]['__text__']
        if 'watchers' in item['data']:
            eitem['watchers'] = item['data']['watchers'][0]['__text__']
        if 'votes' in item['data']:
            eitem['votes'] = item['data']['votes'][0]['__text__']

        if "assigned_to" in issue:
            if "name" in issue["assigned_to"][0]:
                eitem["assigned"] = issue["assigned_to"][0]["name"]
            if "__text__" in issue["assigned_to"][0]:
                eitem["assignee_email"] = issue["assigned_to"][0]["__text__"]


        if "reporter" in issue:
            if "name" in issue["reporter"][0]:
                eitem["reporter_name"] = issue["reporter"][0]["name"]
                eitem["author_name"] = issue["reporter"][0]["name"]
            if "__text__" in issue["reporter"][0]:
                eitem["reporter_email"] = issue["reporter"][0]["__text__"]
                eitem["author_email"] = issue["reporter"][0]["__text__"]

        date_ts = parser.parse(issue['creation_ts'][0]['__text__'])
        eitem['creation_date'] = date_ts.strftime('%Y-%m-%dT%H:%M:%S')


        eitem["bug_id"] = issue['bug_id'][0]['__text__']
        eitem["status"]  = issue['bug_status'][0]['__text__']
        if "short_desc" in issue:
            if "__text__" in issue["short_desc"][0]:
                eitem["main_description"]  = issue['short_desc'][0]['__text__']
        if "summary" in issue:
            if "__text__" in issue["summary"][0]:
                eitem["summary"]  = issue['summary'][0]['__text__']


        # Fix dates
        date_ts = parser.parse(issue['delta_ts'][0]['__text__'])
        eitem['changeddate_date'] = date_ts.isoformat()
        eitem['delta_ts'] = date_ts.strftime('%Y-%m-%dT%H:%M:%S')

        # Add extra JSON fields used in Kibana (enriched fields)
        eitem['comments'] = 0
        eitem['url'] = None

        if 'long_desc' in issue:
            eitem['comments'] = len(issue['long_desc'])
        eitem['url'] = item['origin'] + "/show_bug.cgi?id=" + issue['bug_id'][0]['__text__']
        eitem['resolution_days'] = \
            get_time_diff_days(eitem['creation_date'], eitem['delta_ts'])
        eitem['timeopen_days'] = \
            get_time_diff_days(eitem['creation_date'], datetime.utcnow())

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, self.roles))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(eitem['creation_date'],"bug"))

        return eitem



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Grimoire ELK general utils
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import datetime
from dateutil import parser, tz

def get_time_diff_days(start, end):
    ''' Number of days between two dates in UTC format  '''

    if start is None or end is None:
        return None

    if type(start) is not datetime.datetime:
        start = parser.parse(start).replace(tzinfo=None)
    if type(end) is not datetime.datetime:
        end = parser.parse(end).replace(tzinfo=None)

    seconds_day = float(60*60*24)
    diff_days = \
        (end-start).total_seconds() / seconds_day
    diff_days = float('%.2f' % diff_days)

    return diff_days

# https://github.com/grimoirelab/perceval/blob/master/perceval/utils.py#L149
def unixtime_to_datetime(ut):
    """Convert a unixtime timestamp to a datetime object.
    The function converts a timestamp in Unix format to a
    datetime object. UTC timezone will also be set.
    :param ut: Unix timestamp to convert
    :returns: a datetime object
    :raises InvalidDateError: when the given timestamp cannot be
        converted into a valid date
    """

    dt = datetime.datetime.utcfromtimestamp(ut)
    dt = dt.replace(tzinfo=tz.tzutc())
    return dt



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#
# Copyright (C) 2016 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from datetime import datetime

from dateutil import parser

from .enrich import Enrich

from .utils import get_time_diff_days


class RedmineEnrich(Enrich):

    def get_elastic_mappings(self):

        mapping = """
        {
            "properties": {
                "description_analyzed": {
                  "type": "string",
                  "index":"analyzed"
                  }
           }
        } """

        return {"items":mapping}

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        data = item['data']
        if 'assigned_to' in data:
            identities.append(self.get_sh_identity(data, 'assigned_to'))
        identities.append(self.get_sh_identity(data, 'author'))
        # TODO: identities in journals not added yet

        return identities

    def get_field_author(self):
        return "author"

    def get_sh_identity(self, data, rol):
        identity = {}
        identity['email'] = None
        if rol+"_data" in data:
            if 'mail' in data[rol+"_data"]:
                identity['email'] = data[rol+"_data"]['mail']
        identity['username'] = data[rol]['id']
        identity['name'] = data[rol]['name']
        return identity

    def get_item_sh(self, item):
        """ Add sorting hat enrichment fields for the author of the item """

        eitem = {}  # Item enriched

        identity  = self.get_sh_identity(item['data'], 'author')
        eitem = self.get_item_sh_fields(identity, parser.parse(item[self.get_field_date()]))

        return eitem

    def get_rich_item(self, item):
        eitem = {}

        # metadata fields to copy
        copy_fields = ["metadata__updated_on","metadata__timestamp","ocean-unique-id","origin"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None
        # The real data
        ticket = item['data']

        # data fields to copy
        copy_fields = ["subject","description","updated_on",
                       "due_date", "estimated_hours", "done_ratio",
                       "id", "spent_hours", "start_date", "subject",
                       "last_update"]
        for f in copy_fields:
            if f in ticket:
                eitem[f] = ticket[f]
            else:
                eitem[f] = None
        eitem['description_analyzed'] = eitem['description']
        # Fields which names are translated
        map_fields = {"due_date": "estimated_closing_date",
                      "created_on":"creation_date",
                      "closed_on": "closing_date"}
        for fn in map_fields:
            if fn in ticket:
                eitem[map_fields[fn]] = ticket[fn]
        # Common format
        common = ['category', 'fixed_version', 'priority', 'project', 'status',
                  'tracker', 'author', 'assigned_to']
        for f in common:
            if f in ticket:
                eitem[f+'_id'] = ticket[f]['id']
                eitem[f+'_name'] = ticket[f]['name']

        len_fields = ['attachments', 'changesets', 'journals', 'relations']
        for f in len_fields:
            if f in ticket:
                eitem[f] = len(ticket[f])

        if 'parent' in ticket:
            eitem['parent_id'] = ticket['parent']['id']

        eitem['url'] = eitem['origin']+"/issues/"+str(eitem['id'])

        # Time to
        if "closing_date" in eitem:
            eitem['timeopen_days'] = \
                get_time_diff_days(eitem['creation_date'], eitem['closing_date'])
            eitem['timeworking_days'] = \
                get_time_diff_days(eitem['start_date'], eitem['closing_date'])
        else:
            eitem['timeopen_days'] = \
                get_time_diff_days(eitem['creation_date'], datetime.utcnow())
            eitem['timeworking_days'] = \
                get_time_diff_days(eitem['start_date'], datetime.utcnow())


        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "job"))

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        return eitem



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Perceval2Ocean tool
#
# Copyright (C) 2015 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from datetime import datetime
import logging
from os import sys
from time import time, sleep

from grimoire_elk.arthur import feed_backend, enrich_backend

from grimoire_elk.ocean.conf import ConfOcean

from grimoire_elk.utils import get_elastic
from grimoire_elk.utils import get_params_parser, config_logging

def get_params():
    ''' Get params definition from ElasticOcean and from all the backends '''

    parser = get_params_parser()

    args = parser.parse_args()

    return args


def feed_backends(url, clean, debug = False, redis = None):
    ''' Update Ocean for all existing backends '''

    logging.info("Updating all Ocean")
    elastic = get_elastic(url, ConfOcean.get_index(), clean)
    ConfOcean.set_elastic(elastic)
    fetch_cache = False

    q = Queue('update', connection=Redis(redis), async=async_)

    for repo in ConfOcean.get_repos():
        task_feed = q.enqueue(feed_backend, url, clean, fetch_cache,
                              repo['backend_name'], repo['backend_params'],
                              repo['index'], repo['index_enrich'], repo['project'])
        logging.info("Queued job")
        logging.info(task_feed)

def enrich_backends(url, clean, debug = False, redis = None,
                    db_projects_map=None, db_sortinghat=None):
    ''' Enrich all existing indexes '''

    logging.info("Enriching repositories")

    elastic = get_elastic(url, ConfOcean.get_index(), clean)
    ConfOcean.set_elastic(elastic)
    fetch_cache = False

    q = Queue('update', connection=Redis(redis), async=async_)

    for repo in ConfOcean.get_repos():
        enrich_task = q.enqueue(enrich_backend,
                                url, clean,
                                repo['backend_name'], repo['backend_params'],
                                repo['index'], repo['index_enrich'], db_projects_map, db_sortinghat)
        logging.info("Queued job")
        logging.info(enrich_task)

def loop_update(min_update_time, url, clean, debug, redis, enrich=False,
                db_projects_map=None, db_sortinghat=None):

    while True:
        ustart_feed = time()
        feed_backends(url, clean, debug, redis)
        update_time_feed = int(time()-ustart_feed)
        logging.info("Ocean update time: %i minutes" % (update_time_feed/60))
        update_sleep = min_update_time - update_time_feed

        if enrich:
            ustart_enrich = time()
            enrich_backends(url, clean, debug, redis, db_projects_map, db_sortinghat)
            update_time_enrich = int(time()-ustart_enrich)
            logging.info("Enrich update time: %i minutes" % (update_time_enrich/60))
            update_sleep = min_update_time -(update_time_feed+update_time_enrich)

        if update_sleep > 0:
            logging.debug("Waiting for updating %i seconds " % (update_sleep))
            sleep(update_sleep)

if __name__ == '__main__':

    app_init = datetime.now()

    args = get_params()

    async_ = False  # Use RQ or not

    config_logging(args.debug)

    url = args.elastic_url

    clean = args.no_incremental
    if args.fetch_cache:
        clean = True

    try:
        if args.backend:
            if not args.enrich_only:
                feed_backend(url, clean, args.fetch_cache,
                             args.backend, args.backend_args,
                             args.index, args.index_enrich, args.project)
                logging.info("Backed feed completed")

            if args.enrich or args.enrich_only:
                enrich_backend(url, clean, args.backend, args.backend_args,
                               args.index, args.index_enrich,
                               args.db_projects_map, args.json_projects_map,
                               args.db_sortinghat,
                               args.no_incremental, args.only_identities,
                               args.github_token,
                               args.studies, args.only_studies,
                               args.elastic_url_enrich, args.events_enrich,
                               args.db_user, args.db_password, args.db_host,
                               args.refresh_projects, args.refresh_identities,
                               args.author_id, args.author_uuid, args.filter_raw)
                logging.info("Enrich backend completed")
            elif args.events_enrich:
                logging.info("Enrich option is needed for events_enrich")
        else:
            logging.error("You must configure a backend")

    except KeyboardInterrupt:
        logging.info("\n\nReceived Ctrl-C or other break signal. Exiting.\n")
        sys.exit(0)


    total_time_min = (datetime.now()-app_init).total_seconds()/60

    logging.info("Finished in %.2f min" % (total_time_min))
